import gymnasium as gym
import numpy as np
from pytest import approx

from bsk_rl import act, obs, sats
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_orbit


class TestRelProperties:
    class BasicSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Drift()]

    class RelPropertiesSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [
            obs.RelativeProperties(dict(prop="r_DC_N"), chief_name="Chief")
        ]
        action_spec = [act.Drift()]

    env = gym.make(
        "ConstellationTasking-v1",
        satellites=[
            RelPropertiesSat(
                "Deputy",
                sat_args=dict(
                    oe=random_orbit(i=1.0, a=7000.0, e=0.0, Omega=0.0, omega=0.0, f=0.0)
                ),
            ),
            BasicSat(
                "Chief",
                sat_args=dict(
                    oe=random_orbit(i=1.0, a=7100.0, e=0.0, Omega=0.0, omega=0.0, f=0.0)
                ),
            ),
        ],
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
    )

    def test_relative_observation(self):
        observation, info = self.env.reset()
        assert np.linalg.norm(observation["Deputy"]) == approx(100.0 * 1e3)



"""The ResourceReward system tracks the change in an arbitrary resource level (e.g. fuel, power, time, etc.).

It can be used to penalize the use of resources or reward a gain. The reward type is
configured by setting the ``resource_fn`` argument to the desired function of the satellite.
"""

import logging
from typing import Callable, Union

from bsk_rl.data.base import Data, DataStore, GlobalReward
from bsk_rl.obs import ResourceRewardWeight

logger = logging.getLogger(__name__)


class ResourceData(Data):
    """Data for tracking a change in some resource."""

    def __init__(self, resource_accumulated: float = 0.0) -> None:
        """Construct resource data.

        Args:
            resource_accumulated: Amount of resource change accumulated.
        """
        self.resource_accumulated = resource_accumulated

    def __add__(self, other: "ResourceData") -> "ResourceData":
        """Combine two units of ResourceData.

        Args:
            other: Another unit of ResourceData to add with this one.
        """
        total_resource = self.resource_accumulated + other.resource_accumulated
        return ResourceData(total_resource)

    def __repr__(self) -> str:
        """String representation of the ResourceData."""
        return f"ResourceData(resource_accumulated={self.resource_accumulated})"


class ResourceDataStore(DataStore):
    data_type = ResourceData

    def __init__(self, *args, resource_fn: Callable = lambda sat: 0.0, **kwargs):
        """DataStore for tracking a change in some resource."""
        super().__init__(*args, **kwargs)
        self.resource_fn = resource_fn

    def get_log_state(self) -> float:
        """Gets the current state of the resource."""
        return self.resource_fn(self.satellite)

    def compare_log_states(self, prev_state: float, new_state: float) -> ResourceData:
        """Compare the previous and new state of the resource."""
        delta_resource = new_state - prev_state
        return ResourceData(delta_resource)


class ResourceReward(GlobalReward):
    """Rewards for an arbitrary resource."""

    data_store_type = ResourceDataStore

    def __init__(
        self,
        reward_weight: Union[float, Callable] = 1.0,
        resource_fn: Callable = lambda sat: 0.0,
    ) -> None:
        """Rewards for an arbitrary resource.

        For example, to penalize the use of fuel, set

        .. code-block:: python

            resource_fn = lambda sat: sat.dynamics.dv_available  # or equivalent
            reward_weight = 0.1  # is positive, because the fuel changes are negative


        To penalize the use of time, set

        .. code-block:: python

            resource_fn = lambda sat: sat.simulator.sim_time
            reward_weight = -1e-3  # is negative, because time increases over time

        Args:
            reward_weight: [reward/resource] Scaling factor to apply to changes in resource
                level to yield reward. Can be a float or a function that randomizes the reward
                weight per-episode.
            resource_fn: Function to call to get the resource level for each satellite.
        """
        super().__init__()
        self._reward_weight = reward_weight
        self.resource_fn = resource_fn
        self.data_store_kwargs = dict(resource_fn=resource_fn)

    def reset_pre_sim_init(self) -> None:
        """Reset the reward weight before simulation initialization."""
        if callable(self._reward_weight):
            self.reward_weight = self._reward_weight()
        else:
            self.reward_weight = self._reward_weight
        return super().reset_pre_sim_init()

    def reset_post_sim_init(self) -> None:
        """Add the reward weight to each satellite's observation spec."""
        for satellite in self.scenario.satellites:
            for obs in satellite.observation_builder.observation_spec:
                if isinstance(obs, ResourceRewardWeight):
                    obs.weight_vector.append(self.reward_weight)
        return super().reset_post_sim_init()

    def calculate_reward(
        self, new_data_dict: dict[str, ResourceData]
    ) -> dict[str, float]:
        """Calculate the reward based on the change in resource level times the weight."""
        penalty = {
            sat_name: self.reward_weight * data.resource_accumulated
            for sat_name, data in new_data_dict.items()
        }

        return penalty


__doc_title__ = "Resource Data"
__all__ = ["ResourceData", "ResourceDataStore", "ResourceReward"]



import gymnasium as gym
import pytest

from bsk_rl import act, data, obs, sats, scene
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_orbit


class FullFeaturedDynModel(dyn.GroundStationDynModel, dyn.LOSCommDynModel):
    pass


class TestImagingDynModelStorage:
    @pytest.mark.parametrize(
        "storage_capacity, initial_storage",
        [
            (8e7, -1),
            (8e7, 0),
            (8e7, 8e6),
            (10e7, 10e7),
            (8e7, 10e7),
        ],
    )
    def test_storageInit(self, storage_capacity, initial_storage):
        class ImageSat(sats.ImagingSatellite):
            dyn_type = dyn.ImagingDynModel
            fsw_type = fsw.ImagingFSWModel
            observation_spec = [obs.Time()]
            action_spec = [act.Downlink(), act.Image(n_ahead_image=10)]

        env = gym.make(
            "SatelliteTasking-v1",
            satellite=ImageSat(
                "EO-1",
                sat_args=ImageSat.default_sat_args(
                    oe=random_orbit,
                    dataStorageCapacity=storage_capacity,
                    storageInit=initial_storage,
                ),
            ),
            scenario=scene.UniformTargets(n_targets=1000),
            rewarder=data.NoReward(),
            sim_rate=1.0,
            time_limit=10000.0,
            max_step_duration=1e9,
            disable_env_checker=True,
        )

        env.reset()

        if initial_storage > storage_capacity or initial_storage < 0:
            assert env.unwrapped.satellite.dynamics.storage_level == 0
        else:
            assert env.unwrapped.satellite.dynamics.storage_level == initial_storage

    @pytest.mark.parametrize(
        "storage_capacity, initial_storage",
        [
            (8e7, 8e6),
            (10e7, 10e7),
        ],
    )
    @pytest.mark.flaky(retries=5, delay=1)
    def test_storageInit_downlink(self, storage_capacity, initial_storage):
        class ImageSat(sats.ImagingSatellite):
            dyn_type = FullFeaturedDynModel
            fsw_type = fsw.ImagingFSWModel
            observation_spec = [obs.Time()]
            action_spec = [act.Downlink()]

        env = gym.make(
            "SatelliteTasking-v1",
            satellite=ImageSat(
                "EO-1",
                sat_args=ImageSat.default_sat_args(
                    oe=random_orbit,
                    dataStorageCapacity=storage_capacity,
                    storageInit=initial_storage,
                ),
            ),
            scenario=scene.UniformTargets(n_targets=1000),
            rewarder=data.NoReward(),
            sim_rate=1.0,
            time_limit=10000.0,
            max_step_duration=1e9,
            disable_env_checker=True,
        )

        env.reset()

        terminated = False
        truncated = False
        while not terminated and not truncated:
            observation, reward, terminated, truncated, info = env.step(0)

        assert env.unwrapped.satellite.dynamics.storage_level < initial_storage


class TestConjunctionDynModel:
    @pytest.mark.parametrize(
        "rN1,vN1,collision",
        [
            ([1e8 + 30, 0, 0], [-1, 0, 0], True),
            ([1e8 + 30, 0, 0], [1, 0, 0], False),
            ([0, 1e8, 0], [0, 0, 0], False),
            ([1e8, 0, 0], [0, 0, 0], True),
        ],
    )
    def test_conjunction(self, rN1, vN1, collision):
        class CollisionSat(sats.Satellite):
            fsw_type = fsw.BasicFSWModel
            dyn_type = dyn.ConjunctionDynModel
            observation_spec = [obs.Time()]
            action_spec = [act.Drift()]

        env = gym.make(
            "ConstellationTasking-v1",
            satellites=[
                CollisionSat(
                    "Collision1",
                    sat_args=dict(
                        rN=rN1,
                        vN=vN1,
                        oe=None,
                    ),
                ),
                CollisionSat(
                    "Collision2",
                    sat_args=dict(
                        rN=[1e8, 0, 0],
                        vN=[0, 0, 0],
                        oe=None,
                    ),
                ),
                CollisionSat(
                    "NoCollision3",
                    sat_args=dict(
                        rN=[-1e8, 0, 0],
                        vN=[0, 0, 0],
                        oe=None,
                    ),
                ),
            ],
            sim_rate=1.0,
            time_limit=100.0,
            max_step_duration=1e9,
            disable_env_checker=True,
        )

        env.reset()

        env.step(dict(Collision1=0, Collision2=0, NoCollision3=0))

        sat1 = env.unwrapped.satellites[0]
        sat2 = env.unwrapped.satellites[1]
        sat3 = env.unwrapped.satellites[2]

        if collision:
            assert sat1.dynamics.conjunctions == [sat2]
            assert sat2.dynamics.conjunctions == [sat1]
            assert sat3.dynamics.conjunctions == []
        else:
            assert sat1.dynamics.conjunctions == []
            assert sat2.dynamics.conjunctions == []
            assert sat3.dynamics.conjunctions == []


class TestMaxRangeDynModel:
    @pytest.mark.parametrize(
        "rN1,rN2,max_range_violation",
        [
            ([1e8, 0, 0], [1e8 + 10, 0, 0], False),
            ([1e8, 0, 0], [1e8 + 200, 0, 0], True),
            ([0, 1e8, 0], [1e8, 0, 0], True),
        ],
    )
    @pytest.mark.parametrize("fail_chief", [True, False])
    def test_max_range(self, rN1, rN2, max_range_violation, fail_chief):
        class ChiefSat(sats.Satellite):
            fsw_type = fsw.BasicFSWModel
            dyn_type = dyn.BasicDynamicsModel
            observation_spec = [obs.Time()]
            action_spec = [act.Drift()]

        class DeputySat(sats.Satellite):
            fsw_type = fsw.BasicFSWModel
            dyn_type = dyn.MaxRangeDynModel
            observation_spec = [obs.Time()]
            action_spec = [act.Drift()]

        env = gym.make(
            "ConstellationTasking-v1",
            satellites=[
                ChiefSat(
                    "Chief",
                    sat_args=dict(
                        rN=rN1,
                        vN=[0, 0, 0],
                        oe=None,
                    ),
                ),
                DeputySat(
                    "Deputy",
                    sat_args=dict(
                        rN=rN2,
                        vN=[0, 0, 0],
                        oe=None,
                        chief_name="Chief",
                        max_range_radius=100,
                        enforce_range_on_chief=fail_chief,
                    ),
                ),
            ],
            sim_rate=1.0,
            time_limit=100.0,
            max_step_duration=1e9,
            disable_env_checker=True,
        )

        env.reset()

        _, _, terminated, _, _ = env.step(dict(Chief=0, Deputy=0))

        sat1 = env.unwrapped.satellites[0]
        sat2 = env.unwrapped.satellites[1]

        if max_range_violation:
            assert sat2.dynamics.out_of_ranges == [sat1]
            assert terminated["Deputy"]
            if fail_chief:
                assert terminated["Chief"]
            else:
                assert not terminated["Chief"]
        else:
            assert sat2.dynamics.out_of_ranges == []
            assert not terminated["Chief"]
            assert not terminated["Deputy"]



from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from Basilisk.utilities import macros as mc

from bsk_rl.sim import world
from bsk_rl.sim.dyn import (
    BasicDynamicsModel,
    ContinuousImagingDynModel,
    DynamicsModel,
    GroundStationDynModel,
    ImagingDynModel,
    LOSCommDynModel,
)
from bsk_rl.sim.dyn import base as dyn_module

module = "bsk_rl.sim.dyn."


@patch.multiple(DynamicsModel, __abstractmethods__=set())
class TestDynamicsModel:
    def test_base_class(self):
        sat = MagicMock()
        dyn = DynamicsModel(sat, 1.0)
        dyn.simulator.CreateNewProcess.assert_called_once()
        assert sat.simulator.world == dyn.world
        dyn.reset_for_action()

    @patch(module + "base.check_aliveness_checkers", MagicMock(return_value=True))
    def test_is_alive(self):
        dyn = DynamicsModel(MagicMock(), 1.0)
        assert dyn.is_alive()


basicdyn = module + "BasicDynamicsModel."


def test_basic_requires_world():
    assert world.BasicWorldModel in BasicDynamicsModel._requires_world()


@patch(basicdyn + "_requires_world", MagicMock(return_value=[]))
@patch(basicdyn + "setup_spacecraft_hub")
@patch(basicdyn + "setup_drag_effector")
@patch(basicdyn + "setup_reaction_wheel_dyn_effector")
@patch(basicdyn + "setup_thruster_dyn_effector")
@patch(basicdyn + "setup_simple_nav_object")
@patch(basicdyn + "setup_eclipse_object")
@patch(basicdyn + "setup_solar_panel")
@patch(basicdyn + "setup_battery")
@patch(basicdyn + "setup_power_sink")
@patch(basicdyn + "setup_reaction_wheel_power")
@patch(basicdyn + "setup_thruster_power")
def test_basic_setup_objects(self, *args):
    BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
    for setter in args:
        setter.assert_called_once()


@patch(basicdyn + "_requires_world", MagicMock(return_value=[]))
@patch(basicdyn + "_setup_dynamics_objects", MagicMock())
class TestBasicDynamicsModel:
    def test_dynamic_properties(self):
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.simulator.world = MagicMock(PN=np.identity(3), omega_PN_N=np.zeros(3))
        dyn.scObject = MagicMock()
        message = dyn.scObject.scStateOutMsg.read.return_value
        message.sigma_BN = np.zeros(3)
        message.omega_BN_B = np.zeros(3)
        message.r_BN_N = np.zeros(3)
        message.v_BN_N = np.zeros(3)
        assert (dyn.sigma_BN == np.zeros(3)).all()
        assert (dyn.BN == np.identity(3)).all()
        assert (dyn.omega_BN_B == np.zeros(3)).all()
        assert (dyn.BP == np.identity(3)).all()
        assert (dyn.r_BN_N == np.zeros(3)).all()
        assert (dyn.r_BN_P == np.zeros(3)).all()
        assert (dyn.v_BN_N == np.zeros(3)).all()
        assert (dyn.v_BN_P == np.zeros(3)).all()
        assert (dyn.omega_BP_P == np.zeros(3)).all()

    def test_orbit_properties(self):
        with patch.object(dyn_module, "orbitalMotion") as mocked:
            mocked.rv2elem = MagicMock(
                return_value=MagicMock(a=1.0, e=2.0, i=3.0, Omega=4.0, omega=5.0, f=6.0)
            )
            dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
            dyn.mu = 1
            dyn.scObject = MagicMock()
            dyn.scObject.scStateOutMsg.read.return_value = MagicMock(
                r_BN_N=[1, 2, 3], v_BN_N=[4, 5, 6]
            )
            dyn.simulator.time = 0.0

            # Check that call to rv2elem happens only once
            assert dyn.semi_major_axis == 1.0
            mocked.rv2elem.assert_called_once()
            assert dyn.eccentricity == 2.0
            assert dyn.inclination == 3.0
            assert dyn.ascending_node == 4.0
            assert dyn.argument_of_periapsis == 5.0
            assert dyn.true_anomaly == 6.0
            mocked.rv2elem.assert_called_once()

    def test_orbit_properties_no_cache_update(self):
        with patch.object(dyn_module, "orbitalMotion") as mocked:
            mocked.rv2elem = MagicMock(return_value=MagicMock(a=1.0))
            dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
            dyn.mu = 1
            dyn.scObject = MagicMock()
            dyn.scObject.scStateOutMsg.read.return_value = MagicMock()
            dyn.simulator.time = 0.0

            assert dyn.semi_major_axis == 1.0
            mocked.rv2elem = MagicMock(return_value=MagicMock(a=2.0))
            assert dyn.semi_major_axis == 1.0

    def test_orbit_properties_cache_update(self):
        with patch.object(dyn_module, "orbitalMotion") as mocked:
            mocked.rv2elem = MagicMock(return_value=MagicMock(a=1.0))
            dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
            dyn.mu = 1
            dyn.scObject = MagicMock()
            dyn.scObject.scStateOutMsg.read.return_value = MagicMock()
            dyn.simulator.time = 0.0

            assert dyn.semi_major_axis == 1.0
            dyn.simulator.time = 1.0
            mocked.rv2elem = MagicMock(return_value=MagicMock(a=2.0))
            assert dyn.semi_major_axis == 2.0

    def test_battery_properties(self):
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.powerMonitor = MagicMock()
        dyn.powerMonitor.batPowerOutMsg.read.return_value.storageLevel = 50.0
        dyn.powerMonitor.storageCapacity = 100.0
        assert dyn.battery_charge == 50.0
        assert dyn.battery_charge_fraction == 0.5

    def test_wheel_properties(self):
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.rwStateEffector = MagicMock()
        speeds = np.array([10.0, 20.0, 30.0]) * mc.rpm2radsec
        dyn.rwStateEffector.rwSpeedOutMsg.read.return_value.wheelSpeeds = speeds
        dyn.maxWheelSpeed = 100.0
        assert (dyn.wheel_speeds == speeds).all()
        np.testing.assert_allclose(dyn.wheel_speeds_fraction, np.array([0.1, 0.2, 0.3]))

    @pytest.mark.parametrize(
        "vec,valid",
        [([1, 1, 1], False), ([1e9, 0, 1e9], True)],
    )
    def test_altitude_valid(self, vec, valid):
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.min_orbital_radius = 1e6
        dyn.simulator.world = MagicMock()
        dyn.scObject = MagicMock()
        message = dyn.scObject.scStateOutMsg.read.return_value
        message.r_BN_N = vec
        assert dyn.altitude_valid() == valid

    @pytest.mark.parametrize(
        "speeds,valid",
        [
            ([-10, 0.0, 10.0], True),
            ([200.0, 50.0, -50.0], False),
            ([-200.0, 50.0, -50.0], False),
        ],
    )
    def test_rw_speeds_valid(self, speeds, valid):
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.simulator.world = MagicMock()
        dyn.rwStateEffector = MagicMock()
        speeds = np.array(speeds) * mc.rpm2radsec
        dyn.rwStateEffector.rwSpeedOutMsg.read.return_value.wheelSpeeds = speeds
        dyn.maxWheelSpeed = 100.0
        assert dyn.rw_speeds_valid() == valid

    @pytest.mark.parametrize(
        "level,valid",
        [(10, True), (0, False), (-10, False)],
    )
    def test_battery_valid(self, level, valid):
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.powerMonitor = MagicMock()
        dyn.powerMonitor.batPowerOutMsg.read.return_value.storageLevel = level
        dyn.powerMonitor.storageCapacity = 100.0
        assert dyn.battery_valid() == valid

    @patch("Basilisk.simulation.simpleBattery.SimpleBattery", MagicMock())
    @pytest.mark.parametrize(
        "battery_capacity,init_charge,warning",
        [
            (100, 50, False),
            (100, 101, True),
            (100, 100, False),
            (100, 0.0, False),
            (100, -1, True),
        ],
    )
    def test_battery_init_warning(self, battery_capacity, init_charge, warning) -> None:
        dyn = BasicDynamicsModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.solarPanel = MagicMock()
        dyn.logger = MagicMock()
        dyn.setup_battery(
            batteryStorageCapacity=battery_capacity, storedCharge_Init=init_charge
        )
        if warning:
            dyn.logger.warning.assert_called_with(
                f"Battery initial charge {init_charge} incompatible with its capacity {battery_capacity}."
            )
        else:
            dyn.logger.warning.assert_not_called()


class TestLOSCommDynModel:
    losdyn = module + "LOSCommDynModel."

    @patch(losdyn + "_requires_world", MagicMock(return_value=[]))
    @patch(module + "BasicDynamicsModel._setup_dynamics_objects", MagicMock())
    @patch(losdyn + "setup_los_comms")
    def test_setup_objects(self, *args):
        LOSCommDynModel(MagicMock(simulator=MagicMock()), 1.0)
        for setter in args:
            setter.assert_called_once()

    @patch(losdyn + "_requires_world", MagicMock(return_value=[]))
    @patch(losdyn + "_setup_dynamics_objects", MagicMock())
    @patch(module + "relative_motion.spacecraftLocation", MagicMock())
    def test_setup_los_comms(self):
        mock_sim = MagicMock()
        dyn1 = LOSCommDynModel(MagicMock(simulator=mock_sim), 1.0)
        dyn1.scObject = MagicMock()
        dyn1.setup_los_comms(losMaximumRange=-1, priority=1)
        mock_sim.dynamics_list = {1: dyn1}
        mock_sim.AddModelToTask.assert_not_called()
        assert dyn1.los_comms_ids == []

        dyn2 = LOSCommDynModel(MagicMock(simulator=mock_sim), 1.0)
        dyn2.scObject = MagicMock()
        dyn2.setup_los_comms(losMaximumRange=-1, priority=1)
        mock_sim.dynamics_list[2] = dyn2
        assert dyn1.los_comms_ids == [dyn2.satellite.name]
        assert dyn2.los_comms_ids == [dyn1.satellite.name]
        calls = [
            call(dyn1.task_name, dyn1.losComms, ModelPriority=1),
            call(dyn2.task_name, dyn2.losComms, ModelPriority=1),
        ]
        mock_sim.AddModelToTask.assert_has_calls(calls)

        dyn3 = LOSCommDynModel(MagicMock(simulator=mock_sim), 1.0)
        dyn3.scObject = MagicMock()
        dyn3.setup_los_comms(losMaximumRange=-1, priority=1)
        mock_sim.dynamics_list[3] = dyn3
        assert dyn1.los_comms_ids == [dyn2.satellite.name, dyn3.satellite.name]
        assert dyn2.los_comms_ids == [dyn1.satellite.name, dyn3.satellite.name]
        assert dyn3.los_comms_ids == [dyn1.satellite.name, dyn2.satellite.name]
        calls += [call(dyn3.task_name, dyn3.losComms, ModelPriority=1)]
        mock_sim.AddModelToTask.assert_has_calls(calls)


imdyn = module + "ImagingDynModel."


@patch(imdyn + "_requires_world", MagicMock(return_value=[]))
@patch(module + "BasicDynamicsModel._setup_dynamics_objects", MagicMock())
@patch(imdyn + "setup_instrument_power_sink")
@patch(imdyn + "setup_transmitter_power_sink")
@patch(imdyn + "setup_instrument")
@patch(imdyn + "setup_transmitter")
@patch(imdyn + "setup_storage_unit")
@patch(imdyn + "setup_imaging_target")
def test_setup_objects(*args):
    ImagingDynModel(MagicMock(simulator=MagicMock()), 1.0)
    for setter in args:
        setter.assert_called_once()


class FullFeaturedDynModel(GroundStationDynModel, LOSCommDynModel):
    pass


@patch(imdyn + "_requires_world", MagicMock(return_value=[]))
@patch(imdyn + "_setup_dynamics_objects", MagicMock())
class TestImagingDynModel:
    def test_storage_properties(self):
        dyn = ImagingDynModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.storageUnit = MagicMock()
        dyn.storageUnit.storageUnitDataOutMsg.read.return_value.storageLevel = 50.0
        dyn.storageUnit.storageCapacity = 100.0
        assert dyn.storage_level == 50.0
        assert dyn.storage_level_fraction == 0.5

    @pytest.mark.parametrize(
        "level,valid_check,valid",
        [
            (10, True, True),
            (0, True, True),
            (110, True, False),
            (100.001, True, True),
            (10, False, True),
            (110, False, True),
        ],
    )
    def test_data_storage_valid(self, level, valid_check, valid):
        dyn = ImagingDynModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.storageUnit = MagicMock()
        dyn.storageUnitValidCheck = valid_check
        dyn.storageUnit.storageUnitDataOutMsg.read.return_value.storageLevel = level
        dyn.storageUnit.storageCapacity = 100.0
        assert dyn.data_storage_valid() == valid

    @patch(module + "ground_imaging.partitionedStorageUnit", MagicMock())
    @pytest.mark.parametrize(
        "buffers,names,expected",
        [
            (3, None, ["0", "1", "2"]),
            (None, ["a", "b", "c"], ["a", "b", "c"]),
            (3, ["a", "b", "c"], ["a", "b", "c"]),
            (2, ["a", "b", "c"], ValueError),
        ],
    )
    def test_setup_storage_unit(self, buffers, names, expected):
        mock_sim = MagicMock()
        dyn = ImagingDynModel(MagicMock(simulator=mock_sim), 1.0)
        dyn.instrument = MagicMock()
        dyn.transmitter = MagicMock()
        if isinstance(expected, type):
            with pytest.raises(expected):
                dyn.setup_storage_unit(1000, False, 0, buffers, names)
            return
        dyn.setup_storage_unit(1000, False, 0, buffers, names)
        dyn.storageUnit.addPartition.assert_has_calls([call(name) for name in expected])
        dyn.simulator.CreateNewProcess.assert_called_once()

    @patch(
        "Basilisk.simulation.partitionedStorageUnit.PartitionedStorageUnit", MagicMock()
    )
    @pytest.mark.parametrize(
        "storage_capacity,init_storage,warning",
        [
            (100, 50, False),
            (100, 101, True),
            (100, 100, False),
            (100, 0.0, False),
            (100, -1, True),
        ],
    )
    def test_storage_init_warning(
        self, storage_capacity, init_storage, warning
    ) -> None:
        dyn = ImagingDynModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.instrument = MagicMock()
        dyn.transmitter = MagicMock()
        dyn.setup_storage_unit(
            dataStorageCapacity=storage_capacity,
            storageUnitValidCheck=False,
            storageInit=init_storage,
            transmitterNumBuffers=1,
        )
        if warning:
            dyn.logger.warning.assert_called_with(
                f"Initial storage level {init_storage} incompatible with its capacity {storage_capacity}."
            )
        else:
            dyn.logger.warning.assert_not_called()


class TestGroundStationDynModel:
    def test_requires_world(self):
        assert world.GroundStationWorldModel in GroundStationDynModel._requires_world()

    gsdyn = module + "GroundStationDynModel."

    @patch(gsdyn + "_requires_world", MagicMock(return_value=[]))
    @patch(module + "ImagingDynModel._setup_dynamics_objects", MagicMock())
    @patch(gsdyn + "setup_ground_station_locations")
    def test_setup_objects(self, *args):
        GroundStationDynModel(MagicMock(simulator=MagicMock()), 1.0)
        for setter in args:
            setter.assert_called_once()


@patch(imdyn + "_requires_world", MagicMock(return_value=[]))
@patch(imdyn + "_setup_dynamics_objects", MagicMock())
class TestContinuousImagingDynModel:
    def test_storage_properties(self):
        dyn = ContinuousImagingDynModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.storageUnit = MagicMock()
        dyn.storageUnit.storageUnitDataOutMsg.read.return_value.storageLevel = 50.0
        dyn.storageUnit.storageCapacity = 100.0
        assert dyn.storage_level == 50.0
        assert dyn.storage_level_fraction == 0.5

    @patch("Basilisk.simulation.simpleStorageUnit.SimpleStorageUnit", MagicMock())
    @pytest.mark.parametrize(
        "storage_capacity,init_storage,warning",
        [
            (100, 50, False),
            (100, 101, True),
            (100, 100, False),
            (100, 0.0, False),
            (100, -1, True),
        ],
    )
    def test_storage_init_warning(
        self, storage_capacity, init_storage, warning
    ) -> None:
        dyn = ContinuousImagingDynModel(MagicMock(simulator=MagicMock()), 1.0)
        dyn.instrument = MagicMock()
        dyn.transmitter = MagicMock()
        dyn.setup_storage_unit(
            dataStorageCapacity=storage_capacity,
            storageUnitValidCheck=False,
            storageInit=init_storage,
        )
        if warning:
            dyn.logger.warning.assert_called_with(
                f"Initial storage level {init_storage} incompatible with its capacity {storage_capacity}."
            )
        else:
            dyn.logger.warning.assert_not_called()



import gymnasium as gym
import numpy as np
import pytest
from pytest import approx

from bsk_rl import act, data, obs, sats
from bsk_rl.scene import UniformTargets
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_orbit


class TestImagingSatellite:
    class ImageSat(sats.ImagingSatellite):
        dyn_type = dyn.ImagingDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Image(n_ahead_image=10)]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ImageSat(
            "EO-1",
            initial_generation_duration=1000.0,
            generation_duration=100.0,
            sat_args=ImageSat.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.05,
                imageRateErrorRequirement=0.05,
            ),
        ),
        scenario=UniformTargets(n_targets=5000),
        rewarder=data.UniqueImageReward(),
        sim_rate=1.0,
        time_limit=2000.0,
        max_step_duration=500.0,
        disable_env_checker=True,
    )

    @pytest.mark.skip(reason="Causes tests to hang sometimes")
    def test_generate_more_windows(self):
        self.env.reset()
        dts = []
        while self.env.unwrapped.simulator.sim_time < 1500.0:
            observation, reward, terminated, truncated, info = self.env.step(1)
            dts.append(info["d_ts"])
        assert True  # Ensure stepping past initial generation duration works
        assert (np.array(dts) < 500.0).any()  # Ensure variable interval works

    def test_fixed_interval(self):
        self.env.unwrapped.satellite.variable_interval = False
        self.env.reset()
        dts = []
        while self.env.unwrapped.simulator.sim_time < 1500.0:
            observation, reward, terminated, truncated, info = self.env.step(1)
            dts.append(info["d_ts"])
        assert True  # Ensure stepping past initial generation duration works
        for dt in dts:  # Ensure fixed interval works
            assert np.array(dts) == approx(500.0)



from unittest.mock import MagicMock

import numpy as np
import pytest

from bsk_rl.obs import RelativeProperties
from bsk_rl.obs.relative_observations import rso_imaged_regions


class TestRelativeProperties:
    def test_init(self):
        ob = RelativeProperties(
            dict(name="prop1", fn=lambda d, c: 0.0),
            dict(name="prop2", fn=lambda d, c: 0.0, norm=2.0),
            chief_name="chief",
        )
        assert ob.rel_properties[0]["norm"] == 1.0
        assert len(ob.rel_properties) == 2

    @pytest.mark.parametrize("order", [0, 1])
    def test_identify_chief(self, order):
        ob = RelativeProperties(dict(prop="r_DC_N"), chief_name="chief")
        deputy = MagicMock()
        deputy.name = "deputy"
        chief = MagicMock()
        chief.name = "chief"

        if order == 0:
            deputy.simulator.satellites = [deputy, chief]
        elif order == 1:
            deputy.simulator.satellites = [chief, deputy]
        ob.satellite = deputy
        ob.reset_post_sim_init()
        assert ob.chief == chief

    def test_defined_prop(self):
        ob = RelativeProperties(dict(prop="r_DC_N"), chief_name="chief")
        deputy = MagicMock()
        deputy.dynamics.r_BN_N = np.array([1.0, 2.0, 3.0])
        chief = MagicMock()
        chief.dynamics.r_BN_N = np.array([1.0, 2.0, 3.0])
        ob.satellite = deputy
        ob.chief = chief
        assert np.allclose(ob.get_obs()["r_DC_N"], np.array([0.0, 0.0, 0.0]))

    def test_custom_prop(self):
        ob = RelativeProperties(
            dict(name="prop", fn=lambda d, c: d.prop + c.prop), chief_name="chief"
        )
        deputy = MagicMock(prop=1.0)
        chief = MagicMock(prop=2.0)
        ob.satellite = deputy
        ob.chief = chief
        assert ob.get_obs() == {"prop": 3.0}


class TestRSOImagedRegions:
    def test_one_region(self):
        servicer = MagicMock()
        servicer.data_store.data.point_inspect_status = {
            MagicMock(r_PB_B=np.array([1, 0, 0])): True,
            MagicMock(r_PB_B=np.array([1, 0, 0])): False,
        }
        chief = MagicMock()
        obs = rso_imaged_regions(
            servicer, chief, region_centers=np.array([[1, 0, 0]]), frame="chief_body"
        )
        assert np.allclose(obs, [0.5])

    def test_two_regions(self):
        servicer = MagicMock()
        servicer.data_store.data.point_inspect_status = {
            MagicMock(r_PB_B=np.array([1, 0, 0])): True,
            MagicMock(r_PB_B=np.array([0.9, 0.1, 0])): False,
            MagicMock(r_PB_B=np.array([-0.9, 0, 0.3])): True,
        }
        chief = MagicMock()
        obs = rso_imaged_regions(
            servicer,
            chief,
            region_centers=np.array([[1, 0, 0], [-1, 0, 0]]),
            frame="chief_body",
        )
        assert np.allclose(obs, [0.5, 1.0])

    def test_equally_close_regions(self):
        servicer = MagicMock()
        servicer.data_store.data.point_inspect_status = {
            MagicMock(r_PB_B=np.array([0, 0, 0])): True,
        }
        chief = MagicMock()
        obs = rso_imaged_regions(
            servicer,
            chief,
            region_centers=np.array([[1, 0, 0], [-1, 0, 0]]),
            frame="chief_body",
        )
        assert np.allclose(obs, [1.0, 0.0])



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from pytest import approx

from bsk_rl.scene import RSOPoints, SphericalRSO
from bsk_rl.scene.rso_points import logger
from bsk_rl.sim.dyn import RSODynModel, RSOInspectorDynModel

FakeRSODyn = type("FakeRSODyn", (RSODynModel,), {})
FakeInspectorDyn = type("FakeInspectorDyn", (RSOInspectorDynModel,), {})


@patch.multiple(
    RSOPoints,
    __abstractmethods__=set(),
)
class TestRSOPoints:
    def test_reset_no_rsos(self):
        rso = RSOPoints()
        rso.satellites = [MagicMock(dyn_type=FakeRSODyn)]
        logger.warning = MagicMock()
        rso.reset_pre_sim_init()
        logger.warning.assert_called_once()

    def test_reset_too_many_rsos(self):
        rso = RSOPoints()
        rso.satellites = [
            MagicMock(dyn_type=FakeRSODyn),
            MagicMock(dyn_type=FakeRSODyn),
        ]
        with pytest.raises(AssertionError):
            rso.reset_pre_sim_init()

    def test_reset_id_satellites(self):
        rso_sat = MagicMock(dyn_type=FakeRSODyn)
        inspectors = [
            MagicMock(dyn_type=FakeInspectorDyn),
            MagicMock(dyn_type=FakeInspectorDyn),
        ]
        sats = [rso_sat] + inspectors
        rso = RSOPoints()
        rso.satellites = sats
        rso.reset_pre_sim_init()
        assert rso.rso == rso_sat
        assert rso.inspectors == inspectors


class TestSphericalRSO:
    @pytest.mark.parametrize("n_points", [10, 100, 1000])
    @pytest.mark.parametrize("radius", [0.1, 1.0, 5.0])
    def test_generate_points(self, n_points, radius):
        """Test the generation of points on a sphere."""
        rso = SphericalRSO(n_points=n_points, radius=radius)
        points = rso.generate_points()

        assert len(points) == n_points
        for point in points:
            assert np.linalg.norm(point.r_PB_B) == approx(radius)
            assert np.linalg.norm(point.n_B) == approx(1.0)



"""Basilisk flight software models (FSW) are given in ``bsk_rl.sim.fsw``.

Flight software models serve as the interface between the operation of the satellite in
simulation and the Gymnasium environment. While some FSW models add additional
functionality to the satellite, such as imaging instrument control in :class:`ImagingFSWModel`,
others replace the default control laws with a more complex algorithms, such as :class:`SteeringFSWModel`
vis a vis :class:`BasicFSWModel`.

Actions
-------

Each FSW model has a number of actions that can be called to task the satellite. These
actions are decorated with the :func:`~bsk_rl.sim.fsw.action` decorator, which performs
housekeeping tasks before the action is executed. These actions are the primary way to
control the satellite simulation from other parts of the Gymnasium environment.

Properties
----------

The FSW model provides a number of properties for easy access to the satellite state.
These can be accessed directly from the dynamics model instance, or in the observation
via the :class:`~bsk_rl.obs.SatProperties` observation.

Aliveness Checking
------------------

Certain functions in the FSW models are decorated with the :func:`~bsk_rl.utils.functional.aliveness_checker`
decorator. These functions are called at each step to check if the satellite is still
operational, returning true if the satellite is still alive.
"""

from bsk_rl.sim.fsw.base import BasicFSWModel, FSWModel, SteeringFSWModel, Task, action
from bsk_rl.sim.fsw.ground_imaging import (
    ContinuousImagingFSWModel,
    ImagingFSWModel,
    SteeringImagerFSWModel,
)
from bsk_rl.sim.fsw.orbital import MagicOrbitalManeuverFSWModel
from bsk_rl.sim.fsw.rso_inspection import RSOInspectorFSWModel

__doc_title__ = "FSW Sims"
__all__ = [
    "action",
    "FSWModel",
    "BasicFSWModel",
    "ImagingFSWModel",
    "ContinuousImagingFSWModel",
    "SteeringFSWModel",
    "SteeringImagerFSWModel",
    "MagicOrbitalManeuverFSWModel",
    "RSOInspectorFSWModel",
]



"""Data system for recording RSO surface inspection."""

import logging
from typing import TYPE_CHECKING, Optional

import numpy as np

from bsk_rl.data.base import Data, DataStore, GlobalReward
from bsk_rl.sats import Satellite
from bsk_rl.scene.rso_points import RSOPoint, RSOPoints
from bsk_rl.sim.dyn import RSOInspectorDynModel
from bsk_rl.utils import vizard

if TYPE_CHECKING:
    from bsk_rl.sats import Satellite

logger = logging.getLogger(__name__)

RSO = "rso"
INSPECTOR = "inspector"


class RSOInspectionData(Data):
    def __init__(
        self,
        point_inspect_status: Optional[dict[RSOPoint, bool]] = None,
        point_illuminate_status: Optional[dict[RSOPoint, bool]] = None,
    ):
        """Data for recording RSO surface inspection.

        Args:
            point_inspect_status: Dictionary mapping RSO points to inspection status.
            point_illuminate_status: Dictionary mapping RSO points to illumination status.
        """
        if point_inspect_status is None:
            point_inspect_status = {}
        self.point_inspect_status = point_inspect_status
        if point_illuminate_status is None:
            point_illuminate_status = {}
        self.point_illuminate_status = point_illuminate_status

    def __add__(self, other: "RSOInspectionData"):
        """Perform a boolean AND operation between two RSOInspectionData objects."""
        point_inspect_status = {}
        point_inspect_status.update(self.point_inspect_status)
        for point, access in other.point_inspect_status.items():
            if point not in point_inspect_status:
                point_inspect_status[point] = access
            else:
                point_inspect_status[point] = point_inspect_status[point] or access

        point_illuminate_status = {}
        point_illuminate_status.update(self.point_illuminate_status)
        for point, access in other.point_illuminate_status.items():
            if point not in point_illuminate_status:
                point_illuminate_status[point] = access
            else:
                point_illuminate_status[point] = (
                    point_illuminate_status[point] or access
                )

        return RSOInspectionData(point_inspect_status, point_illuminate_status)

    @property
    def num_points_inspected(self):
        """Number of points inspected."""
        return sum(self.point_inspect_status.values())

    @property
    def num_points_illuminated(self):
        """Number of points illuminated."""
        return sum(self.point_illuminate_status.values())

    def __repr__(self) -> str:
        """String representation of the RSOInspectionData."""
        return (
            f"RSOInspectionData(inspected={self.num_points_inspected}, "
            f"num_points_illuminated={self.num_points_illuminated})"
        )


class RSOInspectionDataStore(DataStore):
    data_type = RSOInspectionData

    def __init__(self, *args, **kwargs):
        """DataStore for recording RSO surface inspection."""
        self.data: RSOInspectionData
        super().__init__(*args, **kwargs)
        self.point_access_recorders = []
        self.storage_recorder = None

        if issubclass(self.satellite.dyn_type, RSOInspectorDynModel):
            self.role = INSPECTOR
        else:
            self.role = RSO

    def set_storage_recorder(self, recorder):
        """Set the storage recorder for the satellite."""
        assert self.role == INSPECTOR
        self.storage_recorder = recorder
        self.satellite.simulator.AddModelToTask(
            self.satellite.dynamics.task_name, recorder, ModelPriority=1000
        )

    def add_point_access_recorder(self, recorder):
        """Add a recorder for point access data."""
        assert self.role == INSPECTOR
        self.point_access_recorders.append(recorder)
        self.satellite.simulator.AddModelToTask(
            self.satellite.dynamics.task_name, recorder, ModelPriority=1000
        )

    def clear_recorders(self):
        """Clear all recorders for point access data."""
        if self.storage_recorder:
            self.storage_recorder.clear()
        for recorder in self.point_access_recorders:
            recorder.clear()

    def get_log_state(self) -> Optional[tuple[list[list[bool]], list[list[bool]]]]:
        """Log the storage unit state and point access state for all times in the step."""
        if self.role == RSO:
            return None

        log_len = len(self.storage_recorder.storageLevel)
        if log_len <= 1:
            imaging_req = np.zeros(log_len)
        else:
            imaging_req = np.diff(self.storage_recorder.storageLevel)
            imaging_req = np.concatenate((imaging_req, [imaging_req[-1]]))

        inspected_logs = []
        illuminated_logs = []
        for recorder in self.point_access_recorders:
            inspected = np.logical_and(imaging_req, recorder.hasAccess)
            inspected_logs.append(list(np.array(inspected)))
            illuminated_logs.append(list(np.array(recorder.hasIllumination)))

        self.clear_recorders()

        return inspected_logs, illuminated_logs

    def compare_log_states(self, _, logs) -> RSOInspectionData:
        """Identify point status."""
        if self.role == RSO:
            return RSOInspectionData()

        inspected_logs, illuminated_logs = logs

        point_inspect_status = {}
        for rso_point, log in zip(
            self.data.point_inspect_status.keys(), inspected_logs
        ):
            if any(log):
                point_inspect_status[rso_point] = True

        point_illuminate_status = {}
        for rso_point, log in zip(
            self.data.point_illuminate_status.keys(), illuminated_logs
        ):
            if any(log):
                point_illuminate_status[rso_point] = True

        if len(point_inspect_status) > 0:
            self.satellite.logger.info(
                f"Inspected {len(point_inspect_status)} points this step"
            )

        return RSOInspectionData(point_inspect_status, point_illuminate_status)


class RSOInspectionReward(GlobalReward):
    data_store_type = RSOInspectionDataStore

    def __init__(
        self,
        inspection_reward_scale: float = 1.0,
        completion_bonus: float = 0.0,
        completion_threshold: float = 0.90,
        min_illuminated_for_completion: float = 0.4,
        min_time_for_completion: float = 5700,
        terminate_on_completion: bool = True,
    ):
        """Reward for RSO inspection.

        A per step reward of ``inspection_reward_scale * (number of points inspected / total points)``
        is yielded at each step. If the number of inspected points inspected exceeds or is equal to
        ``completion_threshold * number of illuminated points`` when either ``min_time_for_completion``
        seconds have passed or at least ``min_illuminated_for_completion`` fraction of points are illuminated,
        the ``completion_bonus`` is awarded.

        Args:
            inspection_reward_scale: Scale factor for the reward based on the number of points inspected.
            completion_bonus: Bonus reward for completing the inspection.
            completion_threshold: Fraction of illuminated points that must be inspected to complete the task.
            min_illuminated_for_completion: Minimum fraction of illuminated points required for completion.
            min_time_for_completion: Minimum simulation time required for completion.
            terminate_on_completion: Whether to terminate the episode when the completion bonus is awarded.
        """
        self.scenario: RSOPoints
        super().__init__()
        self.completion_bonus = completion_bonus
        self.inspection_reward_scale = inspection_reward_scale
        self.completion_threshold = completion_threshold
        self.min_illuminated_for_completion = min_illuminated_for_completion
        self.min_time_for_completion = min_time_for_completion
        self.terminate_on_completion = terminate_on_completion

    def reset_overwrite_previous(self) -> None:
        """Overwrite attributes from previous episode."""
        super().reset_overwrite_previous()
        self.bonus_reward_yielded = False
        self.bonus_reward_time = None

    def reset_post_sim_init(self) -> None:
        """Set up recorders in the data stores."""
        super().reset_post_sim_init()

        for i, inspector in enumerate(self.scenario.inspectors):
            inspector.data_store.set_storage_recorder(
                inspector.dynamics.storageUnit.storageUnitDataOutMsg.recorder()
            )
            logger.debug(
                f"Logging {len(self.scenario.rso.dynamics.rso_points)} access points"
            )
            for rso_point_model in self.scenario.rso.dynamics.rso_points:
                inspector.data_store.add_point_access_recorder(
                    rso_point_model.accessOutMsgs[i].recorder(),
                )

    def initial_data(self, satellite: Satellite) -> RSOInspectionData:
        """Furnish the DataStore with initial data, marking all points as uninspected."""
        if not issubclass(satellite.dyn_type, RSOInspectorDynModel):
            return RSOInspectionData()

        return RSOInspectionData(
            {point: False for point in self.scenario.rso_points},
            {point: False for point in self.scenario.rso_points},
        )

    @vizard.visualize
    def determine_point_colors(self, total_data, new_data_dict):
        """Determine target colors in Vizard."""
        colors = ["grey"] * len(self.scenario.rso_points)
        for i, point in enumerate(self.scenario.rso_points):
            if any(
                [
                    data.point_illuminate_status.get(point, False)
                    for data in new_data_dict.values()
                ]
            ):
                colors[i] = "yellow"
            if total_data.point_inspect_status.get(point, False):
                colors[i] = "chartreuse"

        return colors

    @vizard.visualize
    def update_point_colors(
        self,
        rso_points,
        colors,
        alpha=0.5,
        vizInstance=None,
        vizSupport=None,
    ):
        """Update target colors in Vizard."""
        if not hasattr(self, "prev_colors"):
            self.prev_colors = [None] * len(colors)

        for point, color, prev_color in zip(rso_points, colors, self.prev_colors):
            color_vec = vizSupport.toRGBA255(color, alpha=alpha)
            if prev_color != color:
                vizSupport.changeLocation(vizInstance, str(point), color=color_vec)

        self.prev_colors = colors

    def calculate_reward(
        self, new_data_dict: dict[str, RSOInspectionData]
    ) -> dict[str, float]:
        """Calculate the reward based on the new data generated by each satellite."""
        total_points = len(self.scenario.rso_points)
        reward = {}
        total_data = self.data_type() + self.data
        for satellite_id, data in new_data_dict.items():
            new_points = 0
            for point, access in data.point_inspect_status.items():
                if access and not self.data.point_inspect_status.get(point, False):
                    new_points += 1

            if new_points > 0:
                logger.info(f"{satellite_id} inspected {new_points} new points.")

            reward[satellite_id] = (
                new_points / total_points * self.inspection_reward_scale
            )
            total_data += data

        # Plot inspection status in Vizard
        colors = self.determine_point_colors(total_data, new_data_dict)
        self.update_point_colors(self.scenario.rso_points, colors)

        # Check for completion bonus
        logger.info(
            f"Inspected/Illuminated/Total: {total_data.num_points_inspected}/{total_data.num_points_illuminated}/{total_points}"
        )
        min_illuminated_met = (
            total_data.num_points_illuminated
            >= (len(self.scenario.rso_points) * self.min_illuminated_for_completion)
        ) or (
            self.scenario.satellites[0].simulator.sim_time
            > self.min_time_for_completion
        )

        imaged_fraction_met = (
            total_data.num_points_inspected
            >= self.completion_threshold * total_data.num_points_illuminated
        )

        if (
            min_illuminated_met
            and imaged_fraction_met
            and not self.bonus_reward_yielded
        ):
            logger.info(
                f"{total_data.num_points_inspected} points inspected / {total_data.num_points_illuminated} illuminated! Awarding completion bonus."
            )
            for satellite_id in self.cum_reward:
                reward[satellite_id] = (
                    reward.get(satellite_id, 0.0) + self.completion_bonus
                )
            self.bonus_reward_yielded = True
            self.bonus_reward_time = self.scenario.satellites[0].simulator.sim_time
        return reward

    def is_terminated(self, satellite) -> bool:
        """Terminate the episode if the bonus reward has been yielded."""
        return self.bonus_reward_yielded and self.terminate_on_completion


__doc_title__ = "RSO Inspection"
__all__ = ["RSOInspectionReward", "RSOInspectionDataStore", "RSOInspectionData"]



"""``bsk_rl.utils.rllib.callbacks`` is a collection of utilities for logging additional data in RLlib."""

from typing import TYPE_CHECKING, Callable, Optional, Union

import numpy as np
from deprecated import deprecated
from gymnasium import Wrapper
from pettingzoo.utils import BaseParallelWrapper
from ray.rllib.algorithms.callbacks import DefaultCallbacks

if TYPE_CHECKING:
    from gymnasium import Env
    from pettingzoo import ParallelEnv

    from bsk_rl import GeneralSatelliteTasking
    from bsk_rl.sats import Satellite


@deprecated(
    reason="Only for use with old RLlib stack. Use WrappedEpisodeDataCallbacks instead. This class is not maintained."
)
class EpisodeDataCallbacks(DefaultCallbacks):
    """Deprecated method of logging information at the end of each episode.

    :meta private:
    """

    def __init__(self, *args, **kwargs):
        """Log information at the end of each episode.

        Make a subclass of ``EpisodeDataCallbacks`` and override ``pull_env_metrics`` to
        log environment-specific information at the end of each episode.
        """
        super().__init__(*args, **kwargs)

    def pull_env_metrics(self, env) -> dict[str, float]:
        """Log environment metrics at the end of each episode.

        This function is called whenever ``env`` is terminated or truncated. It should
        return a dictionary of metrics to log.

        Args:
            env: An environment that has completed.
        """
        return {}

    def on_episode_end(
        self,
        env=None,
        metrics_logger=None,
        **kwargs,
    ) -> None:
        """Call pull_env_metrics and log the results.

        :meta private:
        """
        if "base_env" in kwargs:  # Old RLlib stack
            env = kwargs["base_env"].vector_env.envs[0]  # noqa: F841
            env_data = self.pull_env_metrics(env)
            for k, v in env_data.items():
                kwargs["episode"].custom_metrics[k] = v
        else:  # New RLlib stack
            env = env.envs[-1]
            env_data = self.pull_env_metrics(env)
            for k, v in env_data.items():
                metrics_logger.log_value(k, v, clear_on_reduce=True)


@deprecated(
    reason="Use WrappedEpisodeDataCallbacks instead. This class is not maintained."
)
class MultiagentEpisodeDataCallbacks(DefaultCallbacks):
    """Deprecated method of logging information at the end of each episode.

    :meta private:
    """

    def __init__(self, *args, **kwargs):
        """Log information at the end of each episode.

        Make a subclass of :class:`MultiagentEpisodeDataCallbacks` and override
        ``pull_env_metrics`` and ``pull_sat_metrics`` to log environment-specific
        information at the end of each episode. Satellite metrics are logged per-satellite
        and as a mean across all satellites.

        Note that satellites persist in the simulator even after death, so recorded values
        from the end of the episode may not be the same as the values when the agent died.
        """
        super().__init__(*args, **kwargs)

    def pull_env_metrics(self, env) -> dict[str, float]:
        """Log environment metrics at the end of each episode.

        This function is called whenever ``env`` is terminated or truncated. It should
        return a dictionary of metrics to log.

        Args:
            env: An environment that has completed.
        """
        return {}

    def pull_sat_metrics(self, env, satellite) -> dict[str, float]:
        """Log per-satellite metrics at the end of each episode.

        This function is called whenever ``env`` is terminated or truncated. It should
        return a dictionary of metrics to log.

        Args:
            env: An environment that has completed.
            satellite: A satellite in the environment.
        """
        return {}

    def on_episode_end(
        self,
        env=None,
        metrics_logger=None,
        **_,
    ) -> None:
        """Call pull_env_metrics and log the results.

        :meta private:
        """
        env = env.par_env
        env_data = self.pull_env_metrics(env)
        for k, v in env_data.items():
            metrics_logger.log_value(k, v, clear_on_reduce=True)

        all_sat_data = []

        for sat in env.satellites:
            sat_data = self.pull_sat_metrics(env, sat)
            all_sat_data.append(sat_data)
            for k, v in sat_data.items():
                metrics_logger.log_value(f"{sat.id}/{k}", v, clear_on_reduce=True)

        for k in all_sat_data[0].keys():
            metrics_logger.log_value(
                f"mean/{k}",
                np.mean([sat_data[k] for sat_data in all_sat_data]),
                clear_on_reduce=True,
            )


class WrappedEpisodeDataCallbacks(DefaultCallbacks):
    def __init__(self, *args, **kwargs):
        """Log information at the end of each episode.

        Logs data from an environment wrapped with :class:`EpisodeDataWrapper` or
        :class:`EpisodeDataParallelWrapper`. See :doc:`/examples/rllib_training` for an
        example of how to use this class.
        """
        super().__init__(*args, **kwargs)

    def on_environment_created(
        self,
        env=None,
        metrics_logger=None,
        **kwargs,
    ):
        """Identify the environment and attach the metrics logger.

        :meta private:
        """
        if "base_env" in kwargs:  # Old RLlib stack
            env = kwargs["base_env"].vector_env.envs[0]  # noqa: F841
        elif hasattr(env, "par_env"):  # New RLlib stack, parallel
            env = env.par_env
        else:  # New RLlib stack, single_agent
            env = env.envs[-1]
        env.set_metrics_logger(metrics_logger)


class EpisodeDataLogger:
    """Base class for logging data at the end of each episode."""

    def __init__(self, episode_data_callback=None, satellite_data_callback=None):
        """Initialize the episode data logger."""
        if episode_data_callback is None:
            episode_data_callback = lambda _: {}
        self.episode_data_callback = episode_data_callback
        if satellite_data_callback is None:
            satellite_data_callback = lambda *_: {}
        self.satellite_data_callback = satellite_data_callback
        self.metrics_logger = None
        self.env: "GeneralSatelliteTasking"
        self.satellites: list["Satellite"]
        self.logs_to_skip = 3  # Hacky way to avoid a log on initial reset

    def set_metrics_logger(self, metrics_logger):
        """Set the metrics logger for this environment."""
        self.metrics_logger = metrics_logger

    def log_data_on_reset(self):
        """Log data at the end of each episode before resetting the environment."""
        if self.metrics_logger is not None:
            episode_data = self.episode_data_callback(self)
            for k, v in episode_data.items():
                self.metrics_logger.log_value(k, v, clear_on_reduce=True)

            all_sat_data = []

            for sat in self.satellites:
                sat_data = self.satellite_data_callback(self, sat)
                all_sat_data.append(sat_data)
                for k, v in sat_data.items():
                    self.metrics_logger.log_value(
                        f"{sat.id}/{k}", v, clear_on_reduce=True
                    )

            for k in all_sat_data[0].keys():
                self.metrics_logger.log_value(
                    f"mean/{k}",
                    np.mean([sat_data[k] for sat_data in all_sat_data]),
                    clear_on_reduce=True,
                )

    def reset(self, **kwargs):
        """Log data before resetting the environment."""
        if self.logs_to_skip > 0:
            self.logs_to_skip -= 1
        else:
            self.log_data_on_reset()
        return self.env.reset(**kwargs)


class EpisodeDataWrapper(EpisodeDataLogger, Wrapper):
    def __init__(
        self,
        env: "Env",
        episode_data_callback: Optional[
            Callable[["GeneralSatelliteTasking"], dict[str, float]]
        ] = None,
        satellite_data_callback: Optional[
            Callable[["GeneralSatelliteTasking", "Satellite"], dict[str, float]]
        ] = None,
    ):
        """Wrapper for logging data at the end of each multiagent episode.

        This wrapper should be used with the :class:`WrappedEpisodeDataCallbacks` in
        RLlib. At the end of each episode, the environment will log data using the provided
        callback functions. See :doc:`/examples/rllib_training` for an example of
        how to use this class.

        Args:
            env: The environment to wrap.
            episode_data_callback: A function that takes the environment as an argument
                and returns a dictionary of episode-level metrics.
            satellite_data_callback: A function that takes the environment and a satellite
                as arguments and returns a dictionary of satellite-level metrics.
        """
        EpisodeDataLogger.__init__(self, episode_data_callback, satellite_data_callback)
        Wrapper.__init__(self, env)


class EpisodeDataParallelWrapper(EpisodeDataLogger, BaseParallelWrapper):
    def __init__(
        self,
        env: "ParallelEnv",
        episode_data_callback: Optional[
            Callable[["GeneralSatelliteTasking"], dict[str, float]]
        ] = None,
        satellite_data_callback: Optional[
            Callable[["GeneralSatelliteTasking", "Satellite"], dict[str, float]]
        ] = None,
    ):
        """Wrapper for logging data at the end of each multiagent episode.

        This wrapper should be used with the :class:`WrappedEpisodeDataCallbacks` in
        RLlib. At the end of each episode, the environment will log data using the provided
        callback functions. See :doc:`/examples/async_multiagent_training` for an example of
        how to use this class.

        Args:
            env: The environment to wrap.
            episode_data_callback: A function that takes the environment as an argument
                and returns a dictionary of episode-level metrics.
            satellite_data_callback: A function that takes the environment and a satellite
                as arguments and returns a dictionary of satellite-level metrics.
        """
        EpisodeDataLogger.__init__(self, episode_data_callback, satellite_data_callback)
        BaseParallelWrapper.__init__(self, env)


__doc_title__ = "RLlib Callbacks"
__all__ = [
    "EpisodeDataLogger",
    "EpisodeDataWrapper",
    "EpisodeDataParallelWrapper",
    "WrappedEpisodeDataCallbacks",
]



"""Scenarios define data available for satellites to collect."""

import logging
from abc import ABC
from typing import TYPE_CHECKING

from bsk_rl.utils.functional import Resetable

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.data.base import Data
    from bsk_rl.sats import Satellite

logger = logging.getLogger(__name__)


class Scenario(ABC, Resetable):
    """Base scenario class."""

    def __init__(self) -> None:
        self.satellites: list["Satellite"]
        self.utc_init: str

    def link_satellites(self, satellites: list["Satellite"]) -> None:
        """Link the environment satellite list to the scenario.

        Args:
            satellites: List of satellites to communicate between.
        """
        self.satellites = satellites


class UniformNadirScanning(Scenario):
    """Defines a nadir target center at the center of the planet."""

    def __init__(self, value_per_second: float = 1.0) -> None:
        """Construct uniform data over the surface of the planet.

        Can be used with :class:`~bsk_rl.data.ScanningTimeReward`.

        Args:
            value_per_second: Reward per second for imaging nadir.
        """
        self.value_per_second = value_per_second


__all__ = []



"""Basic dynamics model for BSK-RL."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np
from Basilisk.simulation import (
    ReactionWheelPower,
    extForceTorque,
    facetDragDynamicEffector,
    simpleBattery,
    simpleNav,
    simplePowerSink,
    simpleSolarPanel,
    spacecraft,
)
from Basilisk.utilities import (
    RigidBodyKinematics,
    macros,
    orbitalMotion,
    unitTestSupport,
)

from bsk_rl.sim import world
from bsk_rl.utils import actuator_primitives as aP
from bsk_rl.utils.attitude import random_tumble
from bsk_rl.utils.functional import (
    aliveness_checker,
    check_aliveness_checkers,
    default_args,
)
from bsk_rl.utils.orbital import random_orbit, rv2HN, rv2omega

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite
    from bsk_rl.sim import Simulator
    from bsk_rl.sim.world import WorldModel


class DynamicsModel(ABC):
    """Abstract Basilisk dynamics model."""

    @classmethod
    def _requires_world(cls) -> list[type["WorldModel"]]:
        """Define minimum :class:`~bsk_rl.sim.world.WorldModel` for compatibility."""
        return []

    def __init__(
        self,
        satellite: "Satellite",
        dyn_rate: float,
        priority: int = 200,
        **kwargs,
    ) -> None:
        """The abstract base dynamics model.

        One DynamicsModel is instantiated for each satellite in the environment each
        time the environment is reset and new simulator is created.

        Args:
            satellite: Satellite represented by this model.
            dyn_rate: [s] Rate of dynamics simulation.
            priority: Model priority.
            kwargs: Passed through to setup functions.
        """
        self.satellite = satellite
        self.logger = self.satellite.logger.getChild(self.__class__.__name__)

        for required in self._requires_world():
            if not issubclass(type(self.simulator.world), required):
                raise TypeError(
                    f"{self.simulator.world} must be a subclass of {required} to "
                    + f"use dynamics model of type {self.__class__}"
                )

        dyn_proc_name = "DynamicsProcess" + self.satellite.name
        self.dyn_proc = self.simulator.CreateNewProcess(dyn_proc_name, priority)
        self.dyn_rate = dyn_rate
        self.task_name = "DynamicsTask" + self.satellite.name
        self.dyn_proc.addTask(
            self.simulator.CreateNewTask(self.task_name, macros.sec2nano(self.dyn_rate))
        )

        # Initialize all modules and write init one-time messages
        self.scObject: spacecraft.Spacecraft
        self._setup_dynamics_objects(**kwargs)

    @property
    def simulator(self) -> "Simulator":
        """Reference to the episode simulator."""
        return self.satellite.simulator

    @property
    def world(self) -> "WorldModel":
        """Reference to the episode world model."""
        return self.simulator.world

    @abstractmethod  # pragma: no cover
    def _setup_dynamics_objects(self, **kwargs) -> None:
        """Caller for all dynamics object initialization."""
        pass

    def is_alive(self, log_failure=False) -> bool:
        """Check if the dynamics model has failed any aliveness requirements.

        Returns:
            ``True`` if the satellite dynamics are still alive.
        """
        return check_aliveness_checkers(self, log_failure=log_failure)

    def reset_for_action(self) -> None:
        """Reset whenever a flight software :class:`~bsk_rl.sim.fsw.action` is called."""
        pass

    @default_args(utc_init="this value will be set by the world model")
    def _utc_init(self, utc_init: float) -> None:
        """Exists so that utc_init is registered as part of sat_args."""
        pass

    def __del__(self):
        """Log when dynamics are deleted."""
        self.logger.debug("Basilisk dynamics deleted")


class BasicDynamicsModel(DynamicsModel):
    """Basic Dynamics model with minimum necessary Basilisk components."""

    @classmethod
    def _requires_world(cls) -> list[type["WorldModel"]]:
        return [world.BasicWorldModel]

    def __init__(self, *args, **kwargs) -> None:
        """A dynamics model with a basic feature set.

        Includes the following:

        * Spacecraft hub physical properties
        * Gravity
        * Constant disturbance torque (defaults to none)
        * Aerodynamic drag
        * Eclipse checking for power generation
        * Reaction wheels
        * Momentum desaturation thrusters
        * Solar panels, battery, and power system

        Args:
            *args: Passed to superclass
            **kwargs: Passed to superclass
        """
        super().__init__(*args, **kwargs)

    @property
    def sigma_BN(self):
        """Body attitude MRP relative to inertial frame."""
        return self.scObject.scStateOutMsg.read().sigma_BN

    @property
    def BN(self):
        """Body relative to inertial frame rotation matrix."""
        return RigidBodyKinematics.MRP2C(self.sigma_BN)

    @property
    def omega_BN_B(self):
        """Body rate relative to inertial frame in body frame [rad/s]."""
        return self.scObject.scStateOutMsg.read().omega_BN_B

    @property
    def BP(self):
        """Body relative to planet freame rotation matrix."""
        return np.matmul(self.BN, self.world.PN.T)

    @property
    def HN(self):
        """Hill frame relative to inertial frame rotation matrix."""
        return rv2HN(self.r_BN_N, self.v_BN_N)

    @property
    def r_BN_N(self):
        """Body position relative to inertial origin in inertial frame [m]."""
        return self.scObject.scStateOutMsg.read().r_BN_N

    @property
    def r_BN_P(self):
        """Body position relative to inertial origin in planet frame [m]."""
        return np.matmul(self.world.PN, self.r_BN_N)

    @property
    def v_BN_N(self):
        """Body velocity relative to inertial origin in inertial frame [m/s]."""
        return self.scObject.scStateOutMsg.read().v_BN_N

    @property
    def v_BN_P(self):
        """Planet-frame derivative of ``r_BN``."""
        omega_NP_P = np.matmul(self.world.PN, -self.world.omega_PN_N)
        return np.matmul(self.world.PN, self.v_BN_N) + np.cross(omega_NP_P, self.r_BN_P)

    @property
    def omega_BP_P(self):
        """Body angular velocity relative to planet frame in plant frame [rad/s]."""
        omega_BN_N = np.matmul(self.BN.T, self.omega_BN_B)
        omega_BP_N = omega_BN_N - self.world.omega_PN_N
        return np.matmul(self.world.PN, omega_BP_N)

    @property
    def omega_BH_H(self):
        """Body angular velocity relative to Hill frame in Hill frame [rad/s]."""
        omega_BN_N = np.matmul(self.BN.T, self.omega_BN_B)
        omega_HN_N = rv2omega(self.r_BN_N, self.v_BN_N)
        omega_BH_N = omega_BN_N - omega_HN_N
        HN = rv2HN(self.r_BN_N, self.v_BN_N)
        return HN @ omega_BH_N

    def _compute_oes(self):
        if not hasattr(self, "_oe_cache_time") or (
            self._oe_cache_time != getattr(self.simulator, "time", None)
        ):
            self._oe_cache = orbitalMotion.rv2elem(
                mu=self.mu, rVec=np.array(self.r_BN_N), vVec=np.array(self.v_BN_N)
            )
            self._oe_cache_time = getattr(self.simulator, "time", None)
        return self._oe_cache

    @property
    def semi_major_axis(self):
        """Semimajor axis of the satellite's orbit [km]."""
        return self._compute_oes().a

    @property
    def eccentricity(self):
        """Eccentricity of the satellite's orbit [-]."""
        return self._compute_oes().e

    @property
    def inclination(self):
        """Inclination of the satellite's orbit [rad]."""
        return self._compute_oes().i

    @property
    def ascending_node(self):
        """Longitude of ascending node of the satellite's orbit [rad]."""
        return self._compute_oes().Omega

    @property
    def argument_of_periapsis(self):
        """Argument of periapsis of the satellite's orbit [rad]."""
        return self._compute_oes().omega

    @property
    def true_anomaly(self):
        """True anomaly of the satellite's orbit [rad]."""
        return self._compute_oes().f

    @property
    def beta_angle(self):
        """Beta angle of the satellite's orbit, between 0 and 2pi [rad].

        The angle between the angular momentum vector and the sun direction vector.
        """
        r_BN_N = self.r_BN_N
        v_BN_N = self.v_BN_N
        h_N = np.cross(r_BN_N, v_BN_N)
        r_SN_N = (
            self.simulator.world.gravFactory.spiceObject.planetStateOutMsgs[
                self.simulator.world.sun_index
            ]
            .read()
            .PositionVector
        )

        beta = np.arccos(
            np.dot(h_N, r_SN_N) / (np.linalg.norm(h_N) * np.linalg.norm(r_SN_N))
        )
        return beta

    @property
    def battery_charge(self):
        """Battery charge [W*s]."""
        return self.powerMonitor.batPowerOutMsg.read().storageLevel

    @property
    def battery_charge_fraction(self):
        """Battery charge as a fraction of capacity."""
        return self.battery_charge / self.powerMonitor.storageCapacity

    @property
    def wheel_speeds(self):
        """Wheel speeds [rad/s]."""
        return np.array(self.rwStateEffector.rwSpeedOutMsg.read().wheelSpeeds)[0:3]

    @property
    def wheel_speeds_fraction(self):
        """Wheel speeds normalized by maximum allowable speed."""
        return self.wheel_speeds / (self.maxWheelSpeed * macros.rpm2radsec)

    def _setup_dynamics_objects(self, **kwargs) -> None:
        self.setup_spacecraft_hub(**kwargs)
        self.setup_drag_effector(**kwargs)
        self.setup_reaction_wheel_dyn_effector(**kwargs)
        self.setup_thruster_dyn_effector()
        self.setup_simple_nav_object()
        self.setup_eclipse_object()
        self.setup_solar_panel(**kwargs)
        self.setup_battery(**kwargs)
        self.setup_power_sink(**kwargs)
        self.setup_reaction_wheel_power(**kwargs)
        self.setup_thruster_power(**kwargs)

    @default_args(
        mass=330,
        width=1.38,
        depth=1.04,
        height=1.58,
        sigma_init=lambda: random_tumble(maxSpinRate=0.0001)[0],
        omega_init=lambda: random_tumble(maxSpinRate=0.0001)[1],
        rN=None,
        vN=None,
        oe=random_orbit,
        mu=orbitalMotion.MU_EARTH * 1e9,
        min_orbital_radius=(orbitalMotion.REQ_EARTH + 200) * 1e3,
    )
    def setup_spacecraft_hub(
        self,
        mass: float,
        width: float,
        depth: float,
        height: float,
        sigma_init: Iterable[float],
        omega_init: Iterable[float],
        oe: Optional[Iterable[float]],
        rN: Optional[Iterable[float]],
        vN: Optional[Iterable[float]],
        mu: float,
        min_orbital_radius: float,
        priority: int = 2000,
        **kwargs,
    ) -> None:
        """Set up the spacecraft hub physical properties and state.

        The hub is assumed to be a uniform-density rectangular prism with the center of
        mass at the center.

        Args:
            mass: [kg] Hub mass.
            width: [m] Hub width.
            depth: [m] Hub depth.
            height: [m] Hub height.
            sigma_init: Initial attitude MRP.
            omega_init: [rad/s] Initial body rate.
            oe: Orbital element tuple of (semimajor axis [km], eccentricity, inclination
                [rad], ascending node [rad], argument of periapsis [rad], initial true
                anomaly [rad]). Either ``oe`` and ``mu`` or ``rN`` and ``vN`` must be
                provided, but not both.
            mu: Gravitational parameter (used only with ``oe``).
            rN: [m] Initial inertial position.
            vN: [m/s] Initial inertial velocity.
            min_orbital_radius: [m] Minimum allowable orbital radius.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.mu = mu
        if rN is not None and vN is not None and oe is None:
            pass
        elif oe is not None and rN is None and vN is None:
            rN, vN = orbitalMotion.elem2rv(mu, oe)
        else:
            raise (KeyError("Orbit is overspecified. Provide either (rN, vN) or oe"))

        self.scObject = spacecraft.Spacecraft()
        self.scObject.ModelTag = self.satellite.name

        Ixx = 1.0 / 12.0 * mass * (width**2.0 + depth**2.0)
        Iyy = 1.0 / 12.0 * mass * (depth**2.0 + height**2.0)
        Izz = 1.0 / 12.0 * mass * (width**2.0 + height**2.0)
        self.I_mat = [Ixx, 0.0, 0.0, 0.0, Iyy, 0.0, 0.0, 0.0, Izz]

        self.scObject.hub.mHub = mass  # kg
        self.scObject.hub.IHubPntBc_B = unitTestSupport.np2EigenMatrix3d(self.I_mat)

        # Set the initial attitude and position
        self.scObject.hub.sigma_BNInit = sigma_init
        self.scObject.hub.omega_BN_BInit = omega_init
        self.scObject.hub.r_CN_NInit = unitTestSupport.np2EigenVectorXd(rN)
        self.scObject.hub.v_CN_NInit = unitTestSupport.np2EigenVectorXd(vN)

        self.simulator.AddModelToTask(
            self.task_name, self.scObject, ModelPriority=priority
        )

        self.min_orbital_radius = min_orbital_radius

        self.setup_gravity_bodies()
        self.setup_disturbance_torque(**kwargs)
        self.setup_density_model()

    def setup_gravity_bodies(self) -> None:
        """Set up gravitational bodies from the :class:`~bsk_rl.sim.world.WorldModel` to included in the simulation."""
        self.scObject.gravField.gravBodies = spacecraft.GravBodyVector(
            list(self.world.gravFactory.gravBodies.values())
        )

    @default_args(disturbance_vector=None)
    def setup_disturbance_torque(
        self, disturbance_vector: Optional[Iterable[float]] = None, **kwargs
    ) -> None:
        """Set up a constant disturbance torque acting on the satellite.

        Args:
            disturbance_vector: [N*m] Constant disturbance torque in the body frame.
            kwargs: Passed to other setup functions.
        """
        if disturbance_vector is None:
            disturbance_vector = np.array([0, 0, 0])
        self.extForceTorqueObject = extForceTorque.ExtForceTorque()
        self.extForceTorqueObject.ModelTag = "DisturbanceTorque"
        self.extForceTorqueObject.extTorquePntB_B = disturbance_vector
        self.scObject.addDynamicEffector(self.extForceTorqueObject)

    def setup_density_model(self) -> None:
        """Set up the atmospheric density model effector."""
        self.world.densityModel.addSpacecraftToModel(self.scObject.scStateOutMsg)

    @default_args(dragCoeff=2.2)
    def setup_drag_effector(
        self,
        width: float,
        depth: float,
        height: float,
        panelArea: float,
        dragCoeff: float,
        priority: int = 999,
        **kwargs,
    ) -> None:
        """Set up the satellite drag effector.

        The drag effector causes aerodynamic forces and torques to act on the satellite.
        For purposes of this model, the satellite is assumed to be a rectangular prism
        with a solar panel on one end.

        Args:
            width: [m] Hub width.
            depth: [m] Hub depth.
            height: [m] Hub height.
            panelArea: [m^2] Solar panel surface area.
            dragCoeff: Drag coefficient.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.dragEffector = facetDragDynamicEffector.FacetDragDynamicEffector()
        self.dragEffector.ModelTag = "FacetDrag"
        #  Set up the geometry of a small satellite, starting w/ bus
        self.dragEffector.addFacet(
            width * depth, dragCoeff, [1, 0, 0], [height / 2, 0.0, 0]
        )
        self.dragEffector.addFacet(
            width * depth, dragCoeff, [-1, 0, 0], [height / 2, 0.0, 0]
        )
        self.dragEffector.addFacet(
            height * width, dragCoeff, [0, 1, 0], [0, depth / 2, 0]
        )
        self.dragEffector.addFacet(
            height * width, dragCoeff, [0, -1, 0], [0, -depth / 2, 0]
        )
        self.dragEffector.addFacet(
            height * depth, dragCoeff, [0, 0, 1], [0, 0, width / 2]
        )
        self.dragEffector.addFacet(
            height * depth, dragCoeff, [0, 0, -1], [0, 0, -width / 2]
        )
        # Add solar panels
        self.dragEffector.addFacet(
            panelArea / 2,
            dragCoeff,
            [0, 1, 0],
            [0, height, 0],
        )
        self.dragEffector.addFacet(
            panelArea / 2,
            dragCoeff,
            [0, -1, 0],
            [0, height, 0],
        )
        self.dragEffector.atmoDensInMsg.subscribeTo(
            self.world.densityModel.envOutMsgs[-1]
        )
        self.scObject.addDynamicEffector(self.dragEffector)

        self.simulator.AddModelToTask(
            self.task_name, self.dragEffector, ModelPriority=priority
        )

    def setup_simple_nav_object(self, priority: int = 1400, **kwargs) -> None:
        """Set up the navigation module.

        Args:
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.simpleNavObject = simpleNav.SimpleNav()
        self.simpleNavObject.ModelTag = "SimpleNav"
        self.simpleNavObject.scStateInMsg.subscribeTo(self.scObject.scStateOutMsg)
        self.simulator.AddModelToTask(
            self.task_name, self.simpleNavObject, ModelPriority=priority
        )

    @aliveness_checker
    def altitude_valid(self) -> bool:
        """Check that satellite has not deorbited.

        Checks if altitude is greater than the configurable minimum orbital radius
        (`min_orbital_radius`). The default value corresponds to 200km above Earth's surface.
        """
        return np.linalg.norm(self.r_BN_N) > self.min_orbital_radius

    @default_args(
        wheelSpeeds=lambda: np.random.uniform(-1500, 1500, 3),
        maxWheelSpeed=np.inf,
        u_max=0.200,
    )
    def setup_reaction_wheel_dyn_effector(
        self,
        wheelSpeeds: Iterable[float],
        maxWheelSpeed: float,
        u_max: float,
        priority: int = 997,
        **kwargs,
    ) -> None:
        """Set the reaction wheel state effector parameters.

        Three reaction wheels modeled on the HR16 wheel are used.

        Args:
            wheelSpeeds: [rpm] Initial speeds of each wheel.
            maxWheelSpeed: [rpm] Failure speed for wheels.
            u_max: [N*m] Maximum torque producible by each wheel.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.maxWheelSpeed = maxWheelSpeed
        self.rwStateEffector, self.rwFactory, _ = aP.balancedHR16Triad(
            useRandom=False,
            wheelSpeeds=wheelSpeeds,
        )
        for RW in self.rwFactory.rwList.values():
            RW.u_max = u_max
        self.rwFactory.addToSpacecraft(
            "ReactionWheels", self.rwStateEffector, self.scObject
        )
        self.simulator.AddModelToTask(
            self.task_name, self.rwStateEffector, ModelPriority=priority
        )

    @aliveness_checker
    def rw_speeds_valid(self) -> bool:
        """Check if any wheel speed exceeds the ``maxWheelSpeed``."""
        valid = all(
            abs(speed) < self.maxWheelSpeed * macros.rpm2radsec
            for speed in self.wheel_speeds
        )
        return valid

    def setup_thruster_dyn_effector(self, priority: int = 996) -> None:
        """Set up the thruster state effector.

        Args:
            priority: Model priority.
        """
        self.thrusterSet, self.thrFactory = aP.idealMonarc1Octet()
        thrModelTag = "ACSThrusterDynamics"
        self.thrFactory.addToSpacecraft(thrModelTag, self.thrusterSet, self.scObject)
        self.simulator.AddModelToTask(
            self.task_name, self.thrusterSet, ModelPriority=priority
        )

    @default_args(thrusterPowerDraw=0.0)
    def setup_thruster_power(
        self, thrusterPowerDraw, priority: int = 899, **kwargs
    ) -> None:
        """Set up the thruster power draw.

        When momentum desaturating using wheels, power is consumed at this rate.

        Args:
            thrusterPowerDraw: [W] Constant power draw desat mode is active.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.thrusterPowerSink = simplePowerSink.SimplePowerSink()
        self.thrusterPowerSink.ModelTag = "thrustPowerSink" + self.satellite.name
        self.thrusterPowerSink.nodePowerOut = thrusterPowerDraw  # Watts
        self.simulator.AddModelToTask(
            self.task_name, self.thrusterPowerSink, ModelPriority=priority
        )
        self.powerMonitor.addPowerNodeToModel(self.thrusterPowerSink.nodePowerOutMsg)

    def setup_eclipse_object(self) -> None:
        """Add the spacecraft to the eclipse module."""
        self.world.eclipseObject.addSpacecraftToModel(self.scObject.scStateOutMsg)
        self.eclipse_index = len(self.world.eclipseObject.eclipseOutMsgs) - 1

    @default_args(
        panelArea=2 * 1.0 * 0.5,
        panelEfficiency=0.20,
        nHat_B=np.array([0, 0, -1]),
    )
    def setup_solar_panel(
        self,
        panelArea: float,
        panelEfficiency: float,
        nHat_B: Iterable[float],
        priority: int = 898,
        **kwargs,
    ) -> None:
        """Set the solar panel parameters for power generation.

        Power generation takes into account panel size and efficiency, the eclipse
        state, and the angle of solar incidence.

        Args:
            panelArea: [m^2] Solar panel area.
            panelEfficiency: Efficiency coefficient of solar to electrical power
                conversion.
            nHat_B: Body-fixed array normal vector.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.solarPanel = simpleSolarPanel.SimpleSolarPanel()
        self.solarPanel.ModelTag = "solarPanel" + self.satellite.name
        self.solarPanel.stateInMsg.subscribeTo(self.scObject.scStateOutMsg)
        self.solarPanel.sunEclipseInMsg.subscribeTo(
            self.world.eclipseObject.eclipseOutMsgs[self.eclipse_index]
        )
        self.solarPanel.sunInMsg.subscribeTo(
            self.world.gravFactory.spiceObject.planetStateOutMsgs[self.world.sun_index]
        )
        self.solarPanel.setPanelParameters(
            unitTestSupport.np2EigenVectorXd(nHat_B),
            panelArea,
            panelEfficiency,
        )
        self.simulator.AddModelToTask(
            self.task_name, self.solarPanel, ModelPriority=priority
        )

    @default_args(
        batteryStorageCapacity=80.0 * 3600.0,
        storedCharge_Init=lambda: np.random.uniform(30.0 * 3600.0, 70.0 * 3600.0),
    )
    def setup_battery(
        self,
        batteryStorageCapacity: float,
        storedCharge_Init: float,
        priority: int = 799,
        **kwargs,
    ) -> None:
        """Set the battery model parameters.

        Args:
            batteryStorageCapacity: [W*s] Maximum battery charge.
            storedCharge_Init: [W*s] Initial battery charge.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.powerMonitor = simpleBattery.SimpleBattery()
        self.powerMonitor.ModelTag = "powerMonitor"
        self.powerMonitor.storageCapacity = batteryStorageCapacity
        if storedCharge_Init > batteryStorageCapacity or storedCharge_Init < 0:
            self.logger.warning(
                f"Battery initial charge {storedCharge_Init} incompatible with its capacity {batteryStorageCapacity}."
            )
        self.powerMonitor.storedCharge_Init = storedCharge_Init
        self.powerMonitor.addPowerNodeToModel(self.solarPanel.nodePowerOutMsg)
        self.simulator.AddModelToTask(
            self.task_name, self.powerMonitor, ModelPriority=priority
        )

    @aliveness_checker
    def battery_valid(self) -> bool:
        """Check if the battery has charge remaining.

        Note that this check is instantaneous. If a satellite runs out of power during a
        environment step but then recharges to have positive power at the end of the step,
        the satellite will still be considered alive.
        """
        return self.battery_charge > 0

    @default_args(basePowerDraw=0.0)
    def setup_power_sink(
        self, basePowerDraw: float, priority: int = 897, **kwargs
    ) -> None:
        """Set the instrument power sink parameters.

        Args:
            basePowerDraw: [W] Baseline satellite power draw. Should be negative.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        if basePowerDraw > 0:
            self.logger.warning("basePowerDraw should probably be zero or negative.")
        self.basePowerSink = simplePowerSink.SimplePowerSink()
        self.basePowerSink.ModelTag = "basePowerSink" + self.satellite.name
        self.basePowerSink.nodePowerOut = basePowerDraw  # Watts
        self.simulator.AddModelToTask(
            self.task_name, self.basePowerSink, ModelPriority=priority
        )
        self.powerMonitor.addPowerNodeToModel(self.basePowerSink.nodePowerOutMsg)
        self.basePowerSink.powerStatus = 1

    @default_args(
        rwBasePower=0.4, rwMechToElecEfficiency=0.0, rwElecToMechEfficiency=0.5
    )
    def setup_reaction_wheel_power(
        self,
        rwBasePower: float,
        rwMechToElecEfficiency: float,
        rwElecToMechEfficiency: float,
        priority: int = 987,
        **kwargs,
    ) -> None:
        """Set the reaction wheel power draw.

        Args:
            rwBasePower: [W] Constant power draw when operational.
            rwMechToElecEfficiency: Efficiency factor to convert mechanical power to
                electrical power.
            rwElecToMechEfficiency: Efficiency factor to convert electrical power to
                mechanical power.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.rwPowerList = []
        for i_device in range(self.rwFactory.getNumOfDevices()):
            powerRW = ReactionWheelPower.ReactionWheelPower()
            powerRW.ModelTag = "rwPower" + str(i_device)
            powerRW.basePowerNeed = rwBasePower
            powerRW.rwStateInMsg.subscribeTo(self.rwStateEffector.rwOutMsgs[i_device])
            powerRW.mechToElecEfficiency = rwMechToElecEfficiency
            powerRW.elecToMechEfficiency = rwElecToMechEfficiency
            self.rwPowerList.append(powerRW)
            self.simulator.AddModelToTask(
                self.task_name, powerRW, ModelPriority=(priority - i_device)
            )
            self.powerMonitor.addPowerNodeToModel(powerRW.nodePowerOutMsg)


__doc_title__ = "Dynamics Base"

__all__ = [
    "DynamicsModel",
    "BasicDynamicsModel",
]



"""A collection of utilities at ``bsk_rl.utils``.

* :ref:`bsk_rl.utils.actuator_primitives`
* :ref:`bsk_rl.utils.attitude`
* :ref:`bsk_rl.utils.functional`
* :ref:`bsk_rl.utils.orbital`
* :ref:`bsk_rl.utils.rllib`
"""

__doc_title__ = "Utilities"



from collections.abc import Iterable
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from pytest import approx

from bsk_rl import sats
from bsk_rl.scene.targets import Target
from bsk_rl.utils.functional import valid_func_name


@patch(
    "bsk_rl.sats.Satellite.__init__",
    MagicMock(),
)
@patch("bsk_rl.utils.orbital.elevation", lambda x, y: y - x)
@patch.multiple(sats.AccessSatellite, __abstractmethods__=set())
class TestAccessSatellite:
    def make_sat(self):
        sat = sats.AccessSatellite(
            "TestSat",
            sat_args={"imageTargetMinimumElevation": 1},
        )
        sat.logger = MagicMock()
        return sat

    def test_add_location_for_access_checking(self):
        sat = self.make_sat()
        sat.locations_for_access_checking = []
        target = MagicMock()
        sat.add_location_for_access_checking(
            object=target, r_LP_P=[0, 0, 0], min_elev=1.0, type="target"
        )
        assert (
            dict(
                object=target,
                target=target,
                r_LP_P=[0, 0, 0],
                min_elev=1.0,
                type="target",
                start_time=0.0,
            )
            in sat.locations_for_access_checking
        )

    @pytest.mark.parametrize("start", [0.0, 100.0])
    @pytest.mark.parametrize("duration", [0.0, 20.0, 500.0])
    @pytest.mark.parametrize("traj_dt", [30.0, 200.0])
    @pytest.mark.parametrize("generation_duration", [60.0, 100.0])
    def test_calculate_windows_duration(
        self, start, duration, traj_dt, generation_duration
    ):
        sat = self.make_sat()
        sat.window_calculation_time = start
        sat.generation_duration = generation_duration
        sat.trajectory = MagicMock(
            dt=traj_dt,
            r_BP_P=MagicMock(
                x=np.linspace(0, start + duration), y=np.linspace(0, start + duration)
            ),
        )
        sat.locations_for_access_checking = []
        sat.calculate_additional_windows(duration)
        if duration == 0.0:
            return
        assert sat.trajectory.extend_to.call_args[0][0] >= start + duration
        assert sat.trajectory.extend_to.call_args[0][0] - start >= traj_dt * 2

    def test_calculate_windows(self):
        tgt = Target("tgt_0", r_LP_P=[0.0, 0.0, 1.0], priority=1.0)
        sat = self.make_sat()
        sat.window_calculation_time = 0.0
        sat.opportunities = []
        sat.access_dist_threshold = 5.0
        sat.trajectory = MagicMock(
            dt=2.0,
            r_BP_P=MagicMock(
                x=np.arange(0, 100, 2),
                y=np.array([[t - 50.0, 0.0, 2.0] for t in np.arange(0, 100, 2)]),
                side_effect=(  # noqa: E731
                    lambda t: (
                        np.array([[ti - 50.0, 0.0, 2.0] for ti in t])
                        if isinstance(t, Iterable)
                        else np.array([t - 50.0, 0.0, 2.0])
                    )
                ),
            ),
        )
        sat.locations_for_access_checking = [
            dict(
                object=tgt,
                type="target",
                min_elev=1.3,
                r_LP_P=tgt.r_LP_P,
                start_time=9.0,
            )
        ]
        sat.calculate_additional_windows(100.0)
        assert tgt in sat.opportunities_dict()
        assert sat.opportunities[0]["window"][0] == approx(
            50 - 0.27762037530835193, abs=1e-2
        )
        assert sat.opportunities[0]["window"][1] == approx(
            50 + 0.27762037530835193, abs=1e-2
        )

    def test_find_elevation_roots(self):
        interp = lambda t: (  # noqa: E731
            np.array([[ti, 0.0, 2.0] for ti in t])
            if isinstance(t, Iterable)
            else np.array([t, 0.0, 2.0])
        )
        loc = np.array([0.0, 0.0, 1.0])
        elev = 1.3
        times = sats.ImagingSatellite._find_elevation_roots(interp, loc, elev, (-1, 1))
        assert len(times) == 2
        assert times[0] == approx(-times[1], abs=1e-5)
        assert times[1] == approx(0.27762037530835193, abs=1e-5)
        times = sats.ImagingSatellite._find_elevation_roots(interp, loc, elev, (0, 1))
        assert len(times) == 1
        assert times[0] == approx(0.27762037530835193, abs=1e-5)

    @pytest.mark.parametrize(
        "location,times,positions,threshold,expected",
        [
            (
                np.array([2.5, 0.0]),
                np.array([0.0, 10.0, 20.0, 30.0]),
                np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [3.0, 0.0]]),
                1.0,
                [(10.0, 30.0)],
            ),
            (
                np.array([2.5, 0.0]),
                np.array([0.0, 10.0, 20.0, 30.0, 40.0]),
                np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [3.0, 0.0], [4.0, 0.0]]),
                1.0,
                [(10.0, 40.0)],
            ),
            (
                np.array([0.5, 0.0]),
                np.array([0.0, 10.0, 20.0, 30.0]),
                np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [3.0, 0.0]]),
                1.0,
                [(0.0, 20.0)],
            ),
            (
                np.array([1.2, 0.0]),
                np.array([0.0, 10.0, 20.0, 30.0]),
                np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [3.0, 0.0]]),
                5.0,
                [(0.0, 30.0)],
            ),
            (
                np.array([2.5, 100.0]),
                np.array([0.0, 10.0, 20.0, 30.0]),
                np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [3.0, 0.0]]),
                1.0,
                [],
            ),
            (
                np.array([-0.1, 0.0]),
                np.array([0.0, 10.0, 20.0, 30.0, 40.0]),
                np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [3.0, 0.0], [-1.0, 0.0]]),
                1.0,
                [(0.0, 10.0), (30.0, 40.0)],
            ),
        ],
    )
    def test_find_candidate_windows(
        self, location, times, positions, threshold, expected
    ):
        assert (
            sats.ImagingSatellite._find_candidate_windows(
                location, times, positions, threshold
            )
            == expected
        )

    @pytest.mark.parametrize(
        "endpoints,candidate_window,computation_window,expected",
        [
            ([2.4, 14.6], (0.0, 20.0), (0.0, 30.0), [(2.4, 14.6)]),
            ([12.4], (0.0, 20.0), (0.0, 30.0), [(0.0, 12.4)]),
            ([12.4], (10.0, 30.0), (0.0, 30.0), [(12.4, 30.0)]),
            ([2.4, 14.6, 18.8], (0.0, 20.0), (0.0, 30.0), [(0.0, 2.4), (14.6, 18.8)]),
            (
                [2.4, 14.6, 18.8, 19.3],
                (0.0, 20.0),
                (0.0, 30.0),
                [(2.4, 14.6), (18.8, 19.3)],
            ),
        ],
    )
    def test_refine_windows(
        self, endpoints, candidate_window, computation_window, expected
    ):
        assert (
            sats.ImagingSatellite._refine_window(
                endpoints, candidate_window, computation_window
            )
            == expected
        )

    @pytest.mark.skip(reason="Disabled by temporary bugfix")
    def test_refine_windows_impossible(self):
        with pytest.raises(ValueError):
            sats.ImagingSatellite._refine_window(
                [1.0, 2.0, 3.0], (0.0, 4.0), (0.5, 3.5)
            )

    tgt0 = Target("tgt_0", r_LP_P=[0.0, 0.0, 0.0], priority=1.0)
    tgt1 = Target("tgt_1", r_LP_P=[0.0, 0.0, 0.0], priority=1.0)
    tgt2 = Target("tgt_2", r_LP_P=[0.0, 0.0, 0.0], priority=1.0)

    @pytest.mark.parametrize(
        "merge_time",
        [None, 10.0],
    )
    @pytest.mark.parametrize(
        "tgt,window,expected_window",
        [
            (tgt0, (13.0, 18.0), (13.0, 18.0)),
            (tgt2, (13.0, 18.0), (13.0, 18.0)),
            (tgt0, (10.0, 18.0), (2.0, 18.0)),  # Check that merging works
        ],
    )
    def test_add_window(self, merge_time, tgt, window, expected_window):
        sat = self.make_sat()
        sat.opportunities = [
            dict(object=self.tgt1, window=(3.0, 8.0), type="target"),
            dict(object=self.tgt0, window=(2.0, 10.0), type="target"),
        ]
        sat._add_window(
            tgt, window, merge_time=merge_time, type="target", r_LP_P=np.zeros(3)
        )
        assert expected_window in sat.opportunities_dict()[tgt]

    opportunities = [
        dict(object="downObj1", window=(10, 20), type="downlink"),
        dict(object="tgtObj1", window=(20, 30), type="target"),
        dict(object="downObj1", window=(30, 40), type="downlink"),
        dict(object="downObj2", window=(35, 45), type="downlink"),
    ]

    def test_upcoming_opportunities(self):
        sat = self.make_sat()
        sat.opportunities = self.opportunities
        sat.simulator = MagicMock(sim_time=25.0)
        assert sat.upcoming_opportunities == self.opportunities[1:4]

    def test_opportunities_dict(self):
        sat = self.make_sat()
        sat.opportunities = self.opportunities
        assert sat.opportunities_dict(types="target") == dict(tgtObj1=[(20, 30)])
        assert sat.opportunities_dict(types=None) == sat.opportunities_dict(
            types=["target", "downlink"]
        )
        assert sat.opportunities_dict(types="downlink", filter=["downObj1"]) == dict(
            downObj2=[(35, 45)]
        )

    def test_upcoming_opportunities_dict(self):
        sat = self.make_sat()
        sat.opportunities = self.opportunities
        sat.simulator = MagicMock(sim_time=35.0)
        assert sat.upcoming_opportunities_dict(types="target") == {}
        assert sat.upcoming_opportunities_dict(
            types=None
        ) == sat.upcoming_opportunities_dict(types=["target", "downlink"])
        assert sat.upcoming_opportunities_dict(
            types="downlink", filter=["downObj2"]
        ) == dict(downObj1=[(30, 40)])

    def test_next_opportunities_dict(self):
        sat = self.make_sat()
        sat.opportunities = self.opportunities
        sat.simulator = MagicMock(sim_time=15.0)
        assert sat.next_opportunities_dict() == dict(
            downObj1=(10, 20), tgtObj1=(20, 30), downObj2=(35, 45)
        )
        assert sat.next_opportunities_dict(types="downlink") == dict(
            downObj1=(10, 20), downObj2=(35, 45)
        )
        assert sat.next_opportunities_dict(filter=["downObj1"]) == dict(
            tgtObj1=(20, 30), downObj2=(35, 45)
        )

    def test_find_next_opportunities(self):
        pass  # Tested in TestImagingSatellite


@patch("bsk_rl.sats.Satellite.__init__")
@patch.multiple(sats.ImagingSatellite, __abstractmethods__=set())
def test_init(mock_init):
    sats.ImagingSatellite(
        "TestSat",
        sat_args={"imageTargetMinimumElevation": 1},
    )
    mock_init.assert_called_once()


@patch(
    "bsk_rl.sats.Satellite.__init__",
    MagicMock(),
)
@patch.multiple(sats.ImagingSatellite, __abstractmethods__=set())
class TestImagingSatellite:
    def make_sat(self):
        sat = sats.ImagingSatellite(
            "TestSat",
            sat_args={"imageTargetMinimumElevation": 1},
        )
        return sat

    @patch("bsk_rl.sats.Satellite.reset_pre_sim_init")
    def test_reset_pre_sim_init(self, mock_reset):
        sat = self.make_sat()
        sat.observation_builder = MagicMock()
        sat.action_builder = MagicMock()
        sat.reset_overwrite_previous()
        targets = [MagicMock()] * 5
        for target in targets:
            sat.add_location_for_access_checking(target, np.ones(3), 1.0, "target")
        sat.sat_args = {}
        sat.reset_pre_sim_init()
        mock_reset.assert_called_once()
        assert sat.sat_args["transmitterNumBuffers"] == 1
        assert len(sat.sat_args["bufferNames"]) == 1

    @pytest.mark.parametrize(
        "gen_duration,time_limit,expected",
        [(None, float("inf"), 0), (None, 100.0, 100.0), (10.0, 100.0, 10.0)],
    )
    @patch("bsk_rl.sats.Satellite.reset_post_sim_init")
    def test_reset_post_sim_init(self, mock_reset, gen_duration, time_limit, expected):
        sat = self.make_sat()
        sat.sat_args = {}
        sat.calculate_additional_windows = MagicMock()
        sat.initial_generation_duration = gen_duration
        sat.simulator = MagicMock(time_limit=time_limit)
        sat.data_store = MagicMock()
        sat.reset_post_sim_init()
        mock_reset.assert_called_once()
        assert sat.initial_generation_duration == expected
        sat.calculate_additional_windows.assert_called_once()

    tgt0 = Target("tgt_0", r_LP_P=[0.0, 0.0, 0.0], priority=1.0)
    tgt1 = Target("tgt_1", r_LP_P=[0.0, 0.0, 0.0], priority=1.0)
    tgt2 = Target("tgt_2", r_LP_P=[0.0, 0.0, 0.0], priority=1.0)
    windows = {
        tgt0: [(0.0, 10.0), (20.0, 30.0), (40.0, 50.0)],
        tgt1: [(10.0, 20.0)],
        tgt2: [(30.0, 40.0)],
    }
    opportunities = [
        dict(object=tgt0, window=(0.0, 10.0), type="target"),
        dict(object=tgt1, window=(10.0, 20.0), type="target"),
        dict(object=tgt0, window=(20.0, 30.0), type="target"),
        dict(object=tgt2, window=(30.0, 40.0), type="target"),
        dict(object=tgt0, window=(40.0, 50.0), type="target"),
    ]

    @patch(
        "bsk_rl.sats.ImagingSatellite._disable_image_event",
        MagicMock(),
    )
    def test_update_image_event_existing(self):
        sat = self.make_sat()
        sat.name = "Sat"
        tgt = MagicMock(name="tgt")
        existing_event = valid_func_name(f"image_{sat.name}_{tgt.id}")
        sat.simulator = MagicMock(
            eventMap={existing_event: MagicMock(eventActive=False)}
        )
        sat._update_image_event(tgt)
        assert sat.simulator.eventMap[existing_event].eventActive is True

    def test_disable_image_event(self):
        sat = self.make_sat()
        sat.simulator = MagicMock(eventMap={"some_image_event": 1})
        sat._image_event_name = "some_image_event"
        sat._disable_image_event()
        sat.simulator.delete_event.assert_called_with("some_image_event")

    def test_disable_image_event_no_event(self):
        sat = self.make_sat()
        sat.simulator = MagicMock(eventMap={"some_event": 1})
        sat._image_event_name = None
        sat._disable_image_event()
        assert not sat.simulator.delete_event.called

    upcoming_targets = [Target(f"tgt_{i}", [0, 0, 0], 1.0) for i in range(20)]

    @pytest.mark.parametrize(
        "query,expected",
        [
            (0, "tgt_0"),
            (3, "tgt_3"),
            (upcoming_targets[2].id, "tgt_2"),
            (upcoming_targets[2], "tgt_2"),
        ],
    )
    def test_parse_target_selection(self, query, expected):
        sat = self.make_sat()
        sat.find_next_opportunities = lambda *args, **kwargs: [
            dict(object=target) for target in self.upcoming_targets[0 : kwargs["n"]]
        ]
        sat.data_store = MagicMock()
        sat.data_store.data.known = self.upcoming_targets
        assert expected == sat.parse_target_selection(query).name

    def test_parse_target_selection_invalid(self):
        sat = self.make_sat()
        sat.upcoming_targets = lambda x: self.upcoming_targets[0:x]
        sat.data_store = MagicMock()
        sat.data_store.data.known = self.upcoming_targets
        with pytest.raises(TypeError):
            sat.parse_target_selection(np.zeros(10))

    def test_task_target_for_imaging(self):
        sat = self.make_sat()
        sat.opportunities = self.opportunities
        sat.name = "Sat"
        sat.fsw = MagicMock()
        sat.simulator = MagicMock(sim_time=35.0)
        sat._update_image_event = MagicMock()
        sat.update_timed_terminal_event = MagicMock()
        sat.logger = MagicMock()
        sat.task_target_for_imaging(self.tgt0)
        sat.fsw.action_image.assert_called_once()
        sat.logger.info.assert_called()
        sat._update_image_event.assert_called_once()
        assert sat._update_image_event.call_args[0][0] == self.tgt0
        sat.update_timed_terminal_event.assert_called_once()
        assert sat.update_timed_terminal_event.call_args[0][0] == 50.0
        assert sat.latest_target == self.tgt0



from unittest.mock import MagicMock

import pytest

from bsk_rl.data.rso_inspection import (
    RSOInspectionData,
    RSOInspectionDataStore,
    RSOInspectionReward,
    RSOInspectorDynModel,
)


class TestRSOInspectionData:
    @pytest.mark.parametrize(
        "a, b, sum",
        [
            (dict(a=True), dict(a=False), dict(a=True)),
            (dict(a=False), dict(a=False), dict(a=False)),
            (dict(), dict(a=False), dict(a=False)),
            (dict(), dict(a=True), dict(a=True)),
            (dict(a=False), dict(), dict(a=False)),
            (dict(a=True), dict(), dict(a=True)),
            (
                dict(a=True, b=False),
                dict(b=True, c=False),
                dict(a=True, b=True, c=False),
            ),
        ],
    )
    def test_add_inspect(self, a, b, sum):
        da = RSOInspectionData(point_inspect_status=a, point_illuminate_status={})
        db = RSOInspectionData(point_inspect_status=b, point_illuminate_status={})
        result = da + db
        assert sum == result.point_inspect_status
        assert result.point_illuminate_status == {}

    @pytest.mark.parametrize(
        "a, b, sum",
        [
            (dict(a=True), dict(a=False), dict(a=True)),
            (dict(a=False), dict(a=False), dict(a=False)),
            (dict(), dict(a=False), dict(a=False)),
            (dict(), dict(a=True), dict(a=True)),
            (dict(a=False), dict(), dict(a=False)),
            (dict(a=True), dict(), dict(a=True)),
            (
                dict(a=True, b=False),
                dict(b=True, c=False),
                dict(a=True, b=True, c=False),
            ),
        ],
    )
    def test_add_illuminate(self, a, b, sum):
        da = RSOInspectionData(point_inspect_status={}, point_illuminate_status=a)
        db = RSOInspectionData(point_inspect_status={}, point_illuminate_status=b)
        result = da + db
        assert sum == result.point_illuminate_status
        assert result.point_inspect_status == {}


FakeInspectorDyn = type("FakeInspectorDyn", (RSOInspectorDynModel,), {})


class TestRSOInspectionDataStore:
    @pytest.mark.parametrize(
        "storage, access, illumination, expected",
        [
            ([10, 10, 10], [[1, 0, 0]], [[1, 1, 0]], [[0, 0, 0]]),
            ([10, 11, 12], [[1, 0, 0]], [[1, 1, 0]], [[1, 0, 0]]),
            ([10], [[1]], [[1]], [[0]]),
            (
                [10, 11, 12],
                [[1, 0, 0], [1, 1, 0]],
                [[1, 1, 0], [1, 1, 1]],
                [[1, 0, 0], [1, 1, 0]],
            ),
        ],
    )
    def test_get_log_state(self, storage, access, illumination, expected):
        ds = RSOInspectionDataStore(MagicMock(dyn_type=FakeInspectorDyn))
        ds.storage_recorder = MagicMock(storageLevel=storage)
        ds.point_access_recorders = [
            MagicMock(hasAccess=acc_i, hasIllumination=ill_i)
            for acc_i, ill_i in zip(access, illumination)
        ]
        inspected, illuminated = ds.get_log_state()
        assert illuminated == illumination
        assert inspected == expected

    def test_compare_log_states(self):
        ds = RSOInspectionDataStore(MagicMock(dyn_type=FakeInspectorDyn))
        ds.data = MagicMock(
            point_inspect_status=dict(a=None, b=None),
            point_illuminate_status=dict(a=None, b=None),
        )
        data = ds.compare_log_states(
            None, ([[1, 0, 0], [0, 0, 0]], [[0, 0, 0], [1, 1, 0]])
        )
        assert data.point_inspect_status == dict(a=True)
        assert data.point_illuminate_status == dict(b=True)


class TestRSOInspectionReward:
    def test_reward(self):
        r = RSOInspectionReward()
        r.scenario = MagicMock(rso_points=list(range(100)))
        r.scenario.satellites[0].simulator.sim_time = 0.0
        r.data = RSOInspectionData()
        reward = r.calculate_reward({"Sat": RSOInspectionData(dict(a=True, b=False))})
        assert reward == {"Sat": 0.01}

    def test_reward_scaled(self):
        r = RSOInspectionReward(inspection_reward_scale=10)
        r.scenario = MagicMock(rso_points=list(range(100)))
        r.scenario.satellites[0].simulator.sim_time = 0.0
        r.data = RSOInspectionData()
        reward = r.calculate_reward({"Sat": RSOInspectionData(dict(a=True, b=False))})
        assert reward == {"Sat": 0.1}

    def test_reward_already_yielded(self):
        r = RSOInspectionReward()
        r.scenario = MagicMock(rso_points=list(range(100)))
        r.scenario.satellites[0].simulator.sim_time = 0.0
        r.data = RSOInspectionData(dict(a=True, b=False))
        reward = r.calculate_reward({"Sat": RSOInspectionData(dict(a=True, b=True))})
        assert reward == {"Sat": 0.01}

    def test_time_completion(self):
        """Meet the condition of inspected > completion_threshold*illuminated at timeout"""
        r = RSOInspectionReward(
            min_time_for_completion=100.0,
            completion_bonus=10.0,
            min_illuminated_for_completion=1.0,
            completion_threshold=0.8,
        )
        r.bonus_reward_yielded = False
        r.cum_reward = {"Sat": 0.0}
        N = 10
        N_illuminated = 5
        N_seen = 4
        r.scenario = MagicMock(rso_points=list(range(N)))
        r.scenario.satellites[0].simulator.sim_time = 101.0
        r.data = RSOInspectionData({n: False for n in range(N)})
        reward = r.calculate_reward(
            {
                "Sat": RSOInspectionData(
                    {n: True for n in range(N_seen)},
                    {n: True for n in range(N_illuminated)},
                )
            }
        )
        assert reward == {"Sat": 10.0 + 0.4}

    def test_percent_completion(self):
        """Meet the condition of inspected > completion_threshold*illuminated over min illumination"""
        r = RSOInspectionReward(
            min_time_for_completion=100.0,
            completion_bonus=10.0,
            min_illuminated_for_completion=0.5,
            completion_threshold=0.8,
        )
        r.bonus_reward_yielded = False
        r.cum_reward = {"Sat": 0.0}
        N = 10
        N_illuminated = 5
        N_seen = 4
        r.scenario = MagicMock(rso_points=list(range(N)))
        r.scenario.satellites[0].simulator.sim_time = 0.0
        r.data = RSOInspectionData({n: False for n in range(N)})
        reward = r.calculate_reward(
            {
                "Sat": RSOInspectionData(
                    {n: True for n in range(N_seen)},
                    {n: True for n in range(N_illuminated)},
                )
            }
        )
        assert reward == {"Sat": 10.0 + 0.4}

    def test_termination(self):
        r = RSOInspectionReward(
            min_time_for_completion=100.0,
            completion_bonus=10.0,
            terminate_on_completion=True,
        )
        r.bonus_reward_yielded = False
        r.cum_reward = {"Sat": 0.0}
        r.scenario = MagicMock(rso_points=list(range(2)))
        r.scenario.satellites[0].simulator.sim_time = 101.0
        r.data = RSOInspectionData(dict(a=False, b=False))
        r.calculate_reward({"Sat": RSOInspectionData(dict(a=True, b=True))})
        assert r.is_terminated(MagicMock())
        assert r.bonus_reward_yielded



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from pytest import approx

from bsk_rl.data.base import Data, DataStore, GlobalReward
from bsk_rl.data.nadir_data import ScanningTime, ScanningTimeReward, ScanningTimeStore
from bsk_rl.data.no_data import NoData, NoDataStore, NoReward
from bsk_rl.data.resource_data import ResourceData, ResourceDataStore, ResourceReward
from bsk_rl.data.unique_image_data import (
    UniqueImageData,
    UniqueImageReward,
    UniqueImageStore,
)


@patch.multiple(DataStore, __abstractmethods__=set())
class TestDataStore:
    def test_init(self):
        # Essentially a smoketest
        DataStore.data_type = MagicMock
        ds = DataStore(MagicMock())
        ds.get_log_state()

    def test_update_from_logs(self):
        # Essentially a smoketest
        DataStore.data_type = MagicMock
        ds = DataStore(MagicMock())
        ds.update_from_logs()
        ds.update_from_logs()

    def test_update_with_communicated_data(self):
        DataStore.data_type = MagicMock
        ds = DataStore(MagicMock())
        ds.data = []
        ds.stage_communicated_data([100])
        ds.update_with_communicated_data()
        assert ds.data == [100]


@patch.multiple(GlobalReward, __abstractmethods__=set())
class TestGlobalReward:
    def test_reset(self):
        GlobalReward.data_store_type = MagicMock()
        dm = GlobalReward()
        dm.reset_overwrite_previous()
        dm.reset_pre_sim_init()
        dm.reset_post_sim_init()
        assert dm.cum_reward == {}

    def test_create_data_store(self):
        sat = MagicMock()
        data_store = MagicMock()
        GlobalReward.data_store_type = MagicMock(return_value=data_store)
        dm = GlobalReward()
        dm.data_store_kwargs = dict(hello="world")
        dm.scenario = MagicMock()
        dm.reset_overwrite_previous()
        dm.reset_pre_sim_init()
        dm.reset_post_sim_init()
        dm.create_data_store(sat)
        assert sat.data_store == data_store
        assert GlobalReward.data_store_type.call_args.kwargs["hello"] == "world"
        assert sat.name in dm.cum_reward

    def test_reward(self):
        dm = GlobalReward()
        dm.reset_overwrite_previous()
        dm.calculate_reward = MagicMock(return_value={"sat": 10.0})
        dm.cum_reward = {"sat": 5.0}
        assert {"sat": 10.0} == dm.reward({"sat": "data"})
        assert dm.cum_reward == {"sat": 15.0}


class TestNoData:
    def test_add(self):
        dat1 = NoData()
        dat2 = NoData()
        dat = dat1 + dat2
        assert isinstance(dat, NoData)


class TestNoDataStore:
    def test_compare_log_states(self):
        ds = NoDataStore(MagicMock())
        assert isinstance(ds.compare_log_states(0, 1), Data)


class TestNoGlobalReward:
    def test_calculate_reward(self):
        dm = NoReward()
        reward = dm.calculate_reward({"sat1": 0, "sat2": 1})
        assert reward == {"sat1": 0.0, "sat2": 0.0}


class TestUniqueImageData:
    def test_identify_duplicates(self):
        dat1 = UniqueImageData([1, 1, 2])
        assert dat1.duplicates == 1

    def test_add_null(self):
        dat1 = UniqueImageData()
        dat2 = UniqueImageData()
        dat = dat1 + dat2
        assert dat.imaged == set()
        assert dat.duplicates == 0

    def test_add_to_null(self):
        dat1 = UniqueImageData(imaged=[1, 2])
        dat2 = UniqueImageData()
        dat = dat1 + dat2
        assert dat.imaged == {1, 2}
        assert dat.duplicates == 0

    def test_add(self):
        dat1 = UniqueImageData(imaged=[1, 2])
        dat2 = UniqueImageData(imaged=[3, 4])
        dat = dat1 + dat2
        assert dat.imaged == {1, 2, 3, 4}
        assert dat.duplicates == 0

    def test_add_duplicates(self):
        dat1 = UniqueImageData(imaged=[1, 2])
        dat2 = UniqueImageData(imaged=[2, 3])
        dat = dat1 + dat2
        assert dat.imaged == {1, 2, 3}
        assert dat.duplicates == 1

    def test_add_duplicates_existing(self):
        dat1 = UniqueImageData(imaged=[1, 2], duplicates=2)
        dat2 = UniqueImageData(imaged=[2, 3], duplicates=3)
        dat = dat1 + dat2
        assert dat.imaged == {1, 2, 3}
        assert dat.duplicates == 6


class TestUniqueImageStore:
    def test_get_log_state(self):
        sat = MagicMock()
        sat.dynamics.storageUnit.storageUnitDataOutMsg.read().storedData = [10]
        ds = UniqueImageStore(sat)
        assert ds.get_log_state() == 10

    @pytest.mark.parametrize(
        "before,after,imaged",
        [
            ([0], [0], False),
            ([1], [1], False),
            ([1], [0], False),
            ([0], [1], True),
        ],
    )
    def test_compare_log_states(self, before, after, imaged):
        sat = MagicMock()
        ds = UniqueImageStore(sat)
        target = MagicMock()
        ds.data.known = [target]
        sat.latest_target = target
        dat = ds.compare_log_states(np.array(before), np.array(after))
        assert len(dat.imaged) == imaged
        if imaged:
            assert target in dat.imaged


class TestUniqueImagingManager:
    def test_calculate_reward(self):
        dm = UniqueImageReward()
        dm.data = UniqueImageData([])
        reward = dm.calculate_reward(
            {
                "sat1": UniqueImageData([MagicMock(priority=0.1)]),
                "sat2": UniqueImageData([MagicMock(priority=0.2)]),
            }
        )
        assert reward == {"sat1": approx(0.1), "sat2": approx(0.2)}

    def test_calculate_reward_existing(self):
        tgt = MagicMock(priority=0.2)
        dm = UniqueImageReward()
        dm.data = UniqueImageData([tgt])
        reward = dm.calculate_reward(
            {
                "sat1": UniqueImageData([MagicMock(priority=0.1)]),
                "sat2": UniqueImageData([tgt]),
            }
        )
        assert reward == {"sat1": approx(0.1), "sat2": 0.0}

    def test_calculate_reward_repeated(self):
        tgt = MagicMock(priority=0.2)
        dm = UniqueImageReward()
        dm.data = UniqueImageData([])
        reward = dm.calculate_reward(
            {
                "sat1": UniqueImageData([tgt]),
                "sat2": UniqueImageData([tgt]),
            }
        )
        assert reward == {"sat1": approx(0.1), "sat2": approx(0.1)}

    def test_calculate_reward_custom_fn(self):
        dm = UniqueImageReward(reward_fn=lambda x: 1 / x)
        dm.data = UniqueImageData([])
        reward = dm.calculate_reward(
            {
                "sat1": UniqueImageData([MagicMock(priority=1)]),
                "sat2": UniqueImageData([MagicMock(priority=2)]),
            }
        )
        assert reward == {"sat1": approx(1.0), "sat2": 0.5}


class TestNadirScanningTimeData:
    def test_add_null(self):
        dat1 = ScanningTime()
        dat2 = ScanningTime()
        dat = dat1 + dat2
        assert dat.scanning_time == 0.0

    def test_add_to_null(self):
        dat1 = ScanningTime(1.0)
        dat2 = ScanningTime()
        dat = dat1 + dat2
        assert dat.scanning_time == 1.0

    def test_add(self):
        dat1 = ScanningTime(1.0)
        dat2 = ScanningTime(3.0)
        dat = dat1 + dat2
        assert dat.scanning_time == 4.0


class TestScanningNadirTimeStore:
    def test_get_log_state(self):
        sat = MagicMock()
        sat.dynamics.storageUnit.storageUnitDataOutMsg.read().storageLevel = 6
        ds = ScanningTimeStore(sat)
        assert ds.get_log_state() == 6.0

    @pytest.mark.parametrize(
        "before,after,new_time",
        [
            (0, 3, 1),
            (3, 6, 1),
            (1, 1, 0),
            (0, 6, 2),
        ],
    )
    def test_compare_log_states(self, before, after, new_time):
        sat = MagicMock()
        ds = ScanningTimeStore(sat)
        sat.dynamics.instrument.nodeBaudRate = 3
        dat = ds.compare_log_states(before, after)
        assert dat.scanning_time == new_time


class TestNadirScanningManager:
    def test_calculate_reward(self):
        dm = ScanningTimeReward()
        dm.scenario = MagicMock()
        dm.data = ScanningTime([])
        dm.scenario.value_per_second = 1.0
        reward = dm.calculate_reward(
            {
                "sat1": ScanningTime(1),
                "sat2": ScanningTime(2),
            }
        )
        assert reward == {"sat1": 1.0, "sat2": 2.0}

    def test_calculate_reward_existing(self):
        dm = ScanningTimeReward()
        dm.scenario = MagicMock()
        dm.data = ScanningTime(1)
        dm.scenario.value_per_second = 1.0
        reward = dm.calculate_reward(
            {
                "sat1": ScanningTime(2),
                "sat2": ScanningTime(3),
            }
        )
        assert reward == {"sat1": 2.0, "sat2": 3.0}

    def test_calculate_reward_custom_fn(self):
        dm = ScanningTimeReward(reward_fn=lambda x: 1 / x)
        dm.data = ScanningTime([])
        reward = dm.calculate_reward(
            {
                "sat1": ScanningTime(2),
                "sat2": ScanningTime(2),
            }
        )
        assert reward == {"sat1": 0.5, "sat2": 0.5}


class TestResourceData:
    def test_add_null(self):
        dat1 = ResourceData()
        dat2 = ResourceData()
        dat = dat1 + dat2
        assert dat.resource_accumulated == 0.0

    def test_add_to_null(self):
        dat1 = ResourceData(1.0)
        dat2 = ResourceData()
        dat = dat1 + dat2
        assert dat.resource_accumulated == 1.0

    def test_add(self):
        dat1 = ResourceData(1.0)
        dat2 = ResourceData(3.0)
        dat = dat1 + dat2
        assert dat.resource_accumulated == 4.0


class TestResourceDataStore:
    def test_get_log_state(self):
        sat = MagicMock()
        sat.resource_level = 6
        ds = ResourceDataStore(sat, resource_fn=lambda sat: sat.resource_level)
        assert ds.get_log_state() == 6.0

    @pytest.mark.repeat(10)
    def test_compare_log_states(self):
        before = np.random.randint(0, 10)
        after = np.random.randint(0, 10)
        delta = after - before
        sat = MagicMock()
        ds = ResourceDataStore(sat)
        dat = ds.compare_log_states(before, after)
        assert dat.resource_accumulated == delta


class TestResourceReward:
    def test_calculate_reward(self):
        dm = ResourceReward(reward_weight=2.0)
        dm.reset_overwrite_previous()
        dm.reset_pre_sim_init()
        reward = dm.calculate_reward(
            {
                "sat1": ResourceData(1.0),
                "sat2": ResourceData(-2.0),
            }
        )
        assert reward == {"sat1": 2.0, "sat2": -4.0}

    def test_calculate_random_reward(self):
        dm = ResourceReward(reward_weight=lambda: 2.0)
        dm.reset_overwrite_previous()
        dm.reset_pre_sim_init()
        reward = dm.calculate_reward(
            {
                "sat1": ResourceData(1.0),
                "sat2": ResourceData(-2.0),
            }
        )
        assert reward == {"sat1": 2.0, "sat2": -4.0}

    def test_read_reward(self):
        dm = ResourceReward(resource_fn=lambda sat: sat.resource_level)
        dm.reset_overwrite_previous()
        dm.reset_pre_sim_init()
        sat = MagicMock()
        dm.create_data_store(sat)
        sat.resource_level = 3.0
        assert sat.data_store.get_log_state() == 3.0



"""Extended Basilisk SimBaseClass for GeneralSatelliteTasking environments."""

import logging
import os
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any

from Basilisk.utilities import SimulationBaseClass
from Basilisk.utilities import macros as mc

from bsk_rl.utils import vizard

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite
    from bsk_rl.sim.world import WorldModel

logger = logging.getLogger(__name__)


class Simulator(SimulationBaseClass.SimBaseClass):
    """Basilisk simulator for GeneralSatelliteTasking environments."""

    def __init__(
        self,
        satellites: list["Satellite"],
        world_type: type["WorldModel"],
        world_args: dict[str, Any],
        sim_rate: float = 1.0,
        max_step_duration: float = 600.0,
        time_limit: float = float("inf"),
    ) -> None:
        """Basilisk simulator for satellite tasking environments.

        The simulator is reconstructed each time the environment :class:`~bsk_rl.GeneralSatelliteTasking.reset`
        is called, generating a fresh Basilisk simulation.

        Args:
            satellites: Satellites to be simulated
            world_type: Type of world model to be constructed
            world_args: Arguments for world model construction
            sim_rate: [s] Rate for model simulation.
            max_step_duration: [s] Maximum time to propagate sim at a step.
            time_limit: [s] Latest time simulation will propagate to.
        """
        super().__init__()
        self.sim_rate = sim_rate
        self.satellites = satellites
        self.max_step_duration = max_step_duration
        self.time_limit = time_limit
        self.logger = logger
        self.use_simple_earth = False

        self.world: WorldModel

        self._set_world(world_type, world_args)

        self.fsw_list = {}
        self.dynamics_list = {}

        for satellite in self.satellites:
            satellite.set_simulator(self)
            self.dynamics_list[satellite.name] = satellite.set_dynamics(self.sim_rate)
            self.fsw_list[satellite.name] = satellite.set_fsw(self.sim_rate)

    def finish_init(self) -> None:
        """Finish simulator initialization."""
        self.set_vizard_epoch()
        if self.use_simple_earth:
            self.make_earth_simple()
        self.InitializeSimulation()
        self.TotalSim.StepUntilStop(0, -1)

    @property
    def sim_time_ns(self) -> int:
        """Simulation time in ns, tied to SimBase integrator."""
        return self.TotalSim.CurrentNanos

    @property
    def sim_time(self) -> float:
        """Simulation time in seconds, tied to SimBase integrator."""
        return self.sim_time_ns * mc.NANO2SEC

    @vizard.visualize
    def setup_vizard(
        self,
        vizard_rate=None,
        use_simple_earth=False,
        vizSupport=None,
        **vizard_settings,
    ):
        """Setup Vizard for visualization."""
        save_path = Path(vizard.VIZARD_PATH)

        viz_proc_name = "VizProcess"
        viz_proc = self.CreateNewProcess(viz_proc_name, priority=400)

        # Define process name, task name and task time-step
        viz_task_name = "viz_task_name"
        if vizard_rate is None:
            vizard_rate = self.sim_rate
        viz_proc.addTask(self.CreateNewTask(viz_task_name, mc.sec2nano(vizard_rate)))

        customizers = ["spriteList", "genericSensorList"]
        list_data = {}
        for customizer in customizers:
            list_data[customizer] = [
                sat.vizard_data.get(customizer, None) for sat in self.satellites
            ]

        # determine save file: if the configured path is a .bin file, use it directly,
        # otherwise create a timestamped file inside the directory
        if save_path.suffix == ".bin":
            saveFile = save_path
        else:
            saveFile = save_path / f"viz_{time()}.bin"
        saveFile.parent.mkdir(parents=True, exist_ok=True)

        self.vizInstance = vizSupport.enableUnityVisualization(
            self,
            viz_task_name,
            scList=[sat.dynamics.scObject for sat in self.satellites],
            **list_data,
            saveFile=str(saveFile),
        )

        self.use_simple_earth = use_simple_earth
        if self.use_simple_earth:
            vizard_settings["atmospheresOff"] = 1
        for key, value in vizard_settings.items():
            setattr(self.vizInstance.settings, key, value)

        vizard.VIZINSTANCE = self.vizInstance

    @vizard.visualize
    def set_vizard_epoch(self, vizInstance=None):
        """Set the Vizard epoch."""
        vizInstance.epochInMsg.subscribeTo(self.world.gravFactory.epochMsg)

    @vizard.visualize
    def make_earth_simple(self, vizInstance=None, vizSupport=None):
        """Make the Earth shader in Vizard lower detail to help viewing ground locations."""
        earth_texture_path = (
            Path(__file__).resolve().parent.parent
            / "_dat"
            / "world.200407.3x5400x2700.jpg"
        )
        vizSupport.createCustomModel(
            vizInstance,
            simBodiesToModify=["earth"],
            modelPath="HI_DEF_SPHERE",
            customTexturePath=str(earth_texture_path),
        )

    def _set_world(
        self, world_type: type["WorldModel"], world_args: dict[str, Any]
    ) -> None:
        """Construct the simulator world model.

        Args:
            world_type: Type of world model to be constructed.
            world_args: Arguments for world model construction, passed to the world
                from the environment.
        """
        self.world = world_type(self, self.sim_rate, **world_args)

    def run(self) -> None:
        """Propagate the simulator.

        Propagates for a duration up to the ``max_step_duration``, stopping if the
        environment time limit is reached or an event is triggered.
        """
        if "max_step_duration" in self.eventMap:
            self.delete_event("max_step_duration")

        step_end_time = self.sim_time + self.max_step_duration

        self.createNewEvent(
            "max_step_duration",
            mc.sec2nano(self.sim_rate),
            True,
            conditionTime=mc.sec2nano(step_end_time),
            actionFunction=lambda sim: sim.logger.info("Max step duration reached"),
            terminal=True,
        )
        self.ConfigureStopTime(mc.sec2nano(min(self.time_limit, 2**31)))
        self.ExecuteSimulation()

    def delete_event(self, event_name) -> None:
        """Remove an event from the event map.

        Makes event checking faster. Due to a performance issue in Basilisk, it is
        necessary to remove created for tasks that are no longer needed (even if it is
        inactive), or else significant time is spent processing the event at each step.
        """
        del self.eventMap[event_name]

    def get_satellite(self, name: str) -> "Satellite":
        """Get a satellite by name.

        Args:
            name: Name of the satellite to retrieve.

        Returns:
            The satellite object with the specified name.
        """
        for sat in self.satellites:
            if sat.name == name:
                return sat
        raise ValueError(f"Satellite with name '{name}' not found.")

    def __del__(self):
        """Log when simulator is deleted."""
        logger.debug("Basilisk simulator deleted")


__all__ = []



"""``bsk_rl.utils.attitude``: Attitude dynamics related utilities."""

import numpy as np


def random_tumble(maxSpinRate: float = 0.001):
    """Generate a spacecraft random tumble with uniformly sampled conditions.

    Args:
        maxSpinRate: [rad/s] Maximum spin rate.

    Returns:
        tuple:
            * **sigma_bn**: [rad] Initial spacecraft attitude.
            * **omega_bn**: [rad/s] Initial spacecraft angular velocity.
    """
    sigma_bn = np.random.uniform(0, 1.0, [3])
    omega_bn = np.random.uniform(-maxSpinRate, maxSpinRate, [3])

    return sigma_bn, omega_bn


__doc_title__ = "Attitude"
__all__ = ["random_tumble"]



"""Data logging and management for nadir scanning."""

from typing import Callable, Optional

from bsk_rl.data.base import Data, DataStore, GlobalReward
from bsk_rl.sats.satellite import Satellite


class ScanningTime(Data):
    """Data for time spent scanning nadir."""

    def __init__(self, scanning_time: float = 0.0) -> None:
        """Data for time spent scanning nadir.

        Time spent scanning nadir ``scanning_time`` is stored in seconds.

        Args:
            scanning_time: [s] Time spent scanning nadir.
        """
        self.scanning_time = scanning_time

    def __add__(self, other: "ScanningTime") -> "ScanningTime":
        """Define the combination of two units of data."""
        scanning_time = self.scanning_time + other.scanning_time

        return self.__class__(scanning_time)


class ScanningTimeStore(DataStore):
    """DataStore for time spent scanning nadir."""

    data_type = ScanningTime

    def __init__(self, *args, **kwargs) -> None:
        """DataStore for time spent scanning nadir.

        Stores the amount of time spent scanning nadir. Calculates new time spent
        scanning based on baud rate of the instrument and the increase in data stored
        in the buffer.
        """
        super().__init__(*args, **kwargs)

    def get_log_state(self) -> float:
        """Return the amount of data currently stored in the storage unit."""
        storage_unit = self.satellite.dynamics.storageUnit.storageUnitDataOutMsg.read()
        stored_amount = storage_unit.storageLevel

        return stored_amount

    def compare_log_states(self, old_state: float, new_state: float) -> "ScanningTime":
        """Generate a unit of data based on change in stored data amount.

        Args:
            old_state: Previous amount of data in the storage unit.
            new_state: Current amount of data in the storage unit.

        Returns:
            Data: Data generated
        """
        instrument_baudrate = self.satellite.dynamics.instrument.nodeBaudRate

        if new_state > old_state:
            data_generated = (new_state - old_state) / instrument_baudrate
        else:
            data_generated = 0.0

        return ScanningTime(scanning_time=data_generated)


class ScanningTimeReward(GlobalReward):
    """GlobalReward for rewarding time spent scanning nadir."""

    data_store_type = ScanningTimeStore  # type of DataStore managed by the GlobalReward

    def __init__(
        self,
        reward_fn: Optional[Callable] = None,
    ) -> None:
        """GlobalReward for rewarding time spent scanning nadir.

        This class should be used with the :class:`~bsk_rl.scene.UniformNadirScanning`
        scenario and a satellite with :class:`~bsk_rl.sim.fsw.ContinuousImagingFSWModel`
        and the :class:`~bsk_rl.act.Scan` action.

        Time is computed based on the amount of data in the satellite's buffer. In the
        basic configuration, this is the amount of time that the :class:`~bsk_rl.act.Scan`
        action is enabled and pointing thresholds are met. However, if other models are
        used to prevent the accumulation of data, the satellite will not be rewarded for
        those times.

        Args:
            reward_fn: Reward as function of time spend pointing nadir. By default,
                is set to the time spent scanning times ``scenario.value_per_second``.
        """
        super().__init__()
        self._reward_fn = reward_fn

    @property
    def reward_fn(self):
        """Function to calculate reward based on time spent scanning nadir.

        :meta private:
        """
        if self._reward_fn is None:
            return lambda t: t * self.scenario.value_per_second
        return self._reward_fn

    def calculate_reward(
        self, new_data_dict: dict[str, "ScanningTime"]
    ) -> dict[str, float]:
        """Calculate reward based on ``reward_fn``."""
        reward = {}
        for sat, scanning_time in new_data_dict.items():
            reward[sat] = self.reward_fn(scanning_time.scanning_time)

        return reward


__doc_title__ = "Nadir Scanning"
__all__ = ["ScanningTimeReward", "ScanningTimeStore", "ScanningTime"]



"""Relative properties between two satellites."""

from typing import Any

import numpy as np
from Basilisk.utilities.RigidBodyKinematics import MRP2C

from bsk_rl.obs import Observation
from bsk_rl.utils.orbital import rv2HN, rv2omega


def r_DC_N(deputy, chief):
    """Relative position of the deputy satellite to the chief satellite in inertial frame."""
    return np.array(deputy.dynamics.r_BN_N) - np.array(chief.dynamics.r_BN_N)


def r_DC_C(deputy, chief):
    """Relative position of the deputy satellite to the chief satellite in chief body frame."""
    raise NotImplementedError()


def r_DC_D(deputy, chief):
    """Relative position of the deputy satellite to the chief satellite in deputy body frame."""
    raise NotImplementedError()


def r_DC_Hc(deputy, chief):
    """Relative position of the deputy satellite to the chief satellite in chief Hill frame."""
    HcN = chief.dynamics.HN
    return HcN @ r_DC_N(deputy, chief)


def r_DC_Hd(deputy, chief):
    """Relative position of the deputy satellite to the chief satellite in deputy Hill frame."""
    raise NotImplementedError()


def v_DC_N(deputy, chief):
    """Relative velocity of the deputy satellite to the chief satellite in inertial frame."""
    return np.array(deputy.dynamics.v_BN_N) - np.array(chief.dynamics.v_BN_N)


def v_DC_C(deputy, chief):
    """Relative velocity of the deputy satellite to the chief satellite in chief body frame."""
    raise NotImplementedError()


def v_DC_D(deputy, chief):
    """Relative velocity of the deputy satellite to the chief satellite in deputy body frame."""
    raise NotImplementedError()


def v_DC_Hc(deputy, chief):
    """Relative velocity of the deputy satellite to the chief satellite in chief Hill frame."""
    HcN = chief.dynamics.HN
    omega_HcN_N = rv2omega(chief.dynamics.r_BN_N, chief.dynamics.v_BN_N)
    return HcN @ (v_DC_N(deputy, chief) - np.cross(omega_HcN_N, r_DC_N(deputy, chief)))


def v_DC_Hd(deputy, chief):
    """Relative velocity of the deputy satellite to the chief satellite in deputy Hill frame."""
    raise NotImplementedError()


def sigma_DC(deputy, chief):
    """Relative attitude of the deputy satellite to the chief satellite."""
    raise NotImplementedError()


def sigma_DHc(deputy, chief):
    """Relative attitude of the deputy satellite to the chief satellite in chief Hill frame."""
    raise NotImplementedError()


def sigma_HdC(deputy, chief):
    """Relative attitude of the deputy satellite Hill frame to the chief satellite."""
    raise NotImplementedError()


def sigma_HdHc(deputy, chief):
    """Relative attitude of the deputy satellite Hill frame to the chief satellite Hill frame."""
    raise NotImplementedError()


# TODO Could probably make some thing that generates these and other relative properties
# (i.e. whether to use body or hill frame for each sat, what
# frame to express in)


def rso_imaged_regions(
    servicer,
    chief,
    region_centers=np.array(
        [[1, 0, 0], [-1, 0, 0], [0, 1, 0], [0, -1, 0], [0, 0, 1], [0, 0, -1]]
    ),
    frame="chief_hill",
):
    """For use with the RSO tasking scenario.

    Returns the fraction of regions imaged by the servicer, where regions are defined by
    the inspection points closest to each region center.

    Args:
        servicer: The servicer satellite.
        chief: The chief satellite.
        region_centers: The centers of the regions to return inspection metrics for.
            It may be useful to set this as :class:`fibonacci_sphere` points.
        frame: The frame to use for the region centers. Can be ``chief_hill`` or
            ``chief_body``.
    """
    assert frame in ["chief_hill", "chief_body"]
    point_inspect_status = servicer.data_store.data.point_inspect_status

    region_centers_C = []
    for region_center in region_centers:
        if frame == "chief_hill":
            HN = rv2HN(chief.dynamics.r_BN_N, chief.dynamics.v_BN_N)
            CN = MRP2C(chief.dynamics.sigma_BN)
            region_center_C = CN @ HN.T @ region_center
        elif frame == "chief_body":
            region_center_C = region_center

        region_centers_C.append(region_center_C)

    imaged = np.zeros(len(region_centers))
    total = np.zeros(len(region_centers))
    for point, inspected in point_inspect_status.items():
        # find the closest region center
        i_closest = np.argmin(np.linalg.norm(region_centers_C - point.r_PB_B, axis=1))
        total[i_closest] += 1
        if inspected:
            imaged[i_closest] += 1

    return imaged / np.maximum(total, 1)


class RelativeProperties(Observation):
    """Add arbitrary properties relative to some other satellite."""

    def __init__(
        self, *rel_properties: dict[str, Any], chief_name: str, name="rel_props"
    ) -> None:
        """Include properties relative to another satellite.

        Within the observation specification for the deputy satellite, this would look like.

        .. code-block:: python

            obs.RelativeProperties(
                dict(prop="r_DC_N", norm=1e3),
                dict(prop="v_DC_N", norm=1e3),
                chief_name="ChiefSat",
            ),

        Args:
            rel_properties: Property specifications. Properties are optionally
                normalized by some factor. Each observation is a dictionary with the keys:

                * ``prop``: Name of a function in :ref:`bsk_rl.obs.relative_observations`.
                * ``norm`` `optional`: Value to normalize property by. Defaults to 1.0.
                * ``name`` `optional`: Name of the observation element. Defaults to the value of ``prop``.
                * ``fn`` `optional`: Alternatively, call a function that takes the deputy (self) and chief (other)
                  as arguments.
            chief_name: Name of the satellite to compare against.
            name: Name of the observation.
        """
        super().__init__(name=name)

        for rel_property in rel_properties:
            for key in rel_property:
                if key not in ["prop", "norm", "name", "fn"]:
                    raise ValueError(f"Invalid property key: {key}")
            if "norm" not in rel_property:
                rel_property["norm"] = 1.0
            if "fn" not in rel_property:
                try:
                    rel_property["fn"] = globals()[rel_property["prop"]]
                except KeyError:
                    raise ValueError(
                        f"Property prop={rel_property['prop']} is not predefined and no `fn` was provided."
                    )
            if "name" not in rel_property:
                rel_property["name"] = rel_property["prop"]
                if rel_property["norm"] != 1.0:
                    rel_property["name"] += "_normd"

        self.rel_properties = rel_properties
        self.chief_name = chief_name

    def reset_post_sim_init(self) -> None:
        """Connect to the chief satellite.

        :meta private:
        """
        try:
            self.chief = [
                sat
                for sat in self.satellite.simulator.satellites
                if sat.name == self.chief_name
            ][0]
        except IndexError:
            raise ValueError(f"Chief satellite {self.chief_name} not found")

    def get_obs(self) -> dict[str, Any]:
        """Return the observation.

        :meta private:
        """
        obs = {}
        for rel_property in self.rel_properties:
            value = rel_property["fn"](self.satellite, self.chief)
            if isinstance(value, list):
                value = np.array(value)
            norm = rel_property["norm"]
            obs[rel_property["name"]] = value / norm
        return obs


__doc_title__ = "Relative Properties"
__all__ = [
    "RelativeProperties",
    "r_DC_N",
    "r_DC_C",
    "r_DC_D",
    "r_DC_Hc",
    "r_DC_Hd",
    "v_DC_N",
    "v_DC_C",
    "v_DC_D",
    "v_DC_Hc",
    "v_DC_Hd",
    "sigma_DC",
    "sigma_DHc",
    "sigma_HdC",
    "sigma_HdHc",
    "rso_imaged_regions",
]



"""Utilities for Vizard visualization."""

import inspect
import logging
from functools import wraps

logger = logging.getLogger(__name__)

VIZARD_PATH = None
VIZINSTANCE = None


def visualize(func):
    """Decorator for functions that enable Vizard."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if VIZARD_PATH is None:
            return

        from Basilisk.simulation import vizInterface
        from Basilisk.utilities import vizSupport

        if not vizSupport.vizFound:
            logger.warning("Vizard not found, disabling visualization")
            return

        sig = inspect.signature(func)

        if "vizInstance" in sig.parameters:
            kwargs["vizInstance"] = VIZINSTANCE
        if "vizSupport" in sig.parameters:
            kwargs["vizSupport"] = vizSupport
        if "vizInterface" in sig.parameters:
            kwargs["vizInterface"] = vizInterface

        return func(*args, **kwargs)

    return wrapper


@visualize
def get_color(index):
    from matplotlib.colors import TABLEAU_COLORS

    n_colors = len(TABLEAU_COLORS)
    color = list(TABLEAU_COLORS.keys())[index % n_colors]
    return color


__doc_title__ = "Vizard"
__all__ = ["visualize", "VIZARD_PATH"]
